"""The "fred/tags" APIs.

See the official FRED API docs for more info:

    https://fred.stlouisfed.org/docs/api/fred/

"""

import pandas as pd

from . import _api


class RelatedTags(_api.API):
    """Get FRED tags related to other FRED tags.

    The module variable :data:`related_tags` is an instance of this API
    implementation and is the most popular interface for calling
    this API.

    """

    url = "https://api.stlouisfed.org/fred/related_tags"

    @classmethod
    def get(
        cls,
        *,
        realtime_start: None | int | str = None,
        realtime_end: None | int | str = None,
        tag_names: None | str | list[str] = None,
        exclude_tag_names: None | str | list[str] = None,
        tag_group_id: None | str = None,
        search_text: None | str | list[str] = None,
        limit: None | int = 1000,
        offset: None | int = 0,
        order_by: None | str = None,
        sort_order: None | str = None,
        paginate: bool = False,
        cache: bool = True,
        api_key: None | str = None,
    ) -> pd.DataFrame:
        """Get data for tags related to an economic release.

        See the related FRED API documentation at:

            https://fred.stlouisfed.org/docs/api/fred/related_tags.html

        Args:
            realtime_start: Start date for fetching results
                according to their publication date.
            realtime_end: End date for fetching results according
                to their publication date.
            tag_names: Find tags related to these tags.
            exclude_tag_names: Exclude tags related to these tags.
            tag_group_id: A tag group ID to filter tags by. Options include:

                - "freq" = frequency
                - "gen" = general or concept
                - "geo" = geography
                - "geot" = geography type
                - "rls" = release
                - "seas" = seasonal adjustment
                - "src" = source

            search_text: The words to find matching tags with.
            limit: Maximum number of results to return.
            offset: Result start offset.
            order_by: Variable to order results by. Options include:

                - "series_count"
                - "popularity"
                - "created"
                - "name"
                - "group_id"

            sort_order: Sort results in ascending ("asc") or
                descending ("desc") order.
            paginate: Whether to manage `offset` automatically, making multiple
                API calls until all results are returned.
            cache: Whether to cache the response from the API.
            api_key: Your FRED API key. Defaults to the ``FRED_API_KEY``
                environment variable.

        Returns:
            A dataframe containing data for related FRED tags.

        Examples:
            >>> finagg.fred.api.related_tags.get(tag_names="bea", limit=5)  # doctest: +SKIP
                                            name group_id                     notes                 created  popularity  series_count
            0  public domain: citation requested       cc                      None  2018-12-17 23:33:13-06          99         78680
            1                                usa      geo  United States of America  2012-02-27 10:18:19-06         100         78434
            2                                nsa     seas   Not Seasonally Adjusted  2012-02-27 10:18:19-06          99         67720
            3                             annual     freq                            2012-02-27 10:18:19-06          88         66478
            4                                gdp      gen    Gross Domestic Product  2012-02-27 10:18:19-06          81         60040

        """
        return _api.maybe_paginate(
            "tags",
            cls.url,
            realtime_start=realtime_start,
            realtime_end=realtime_end,
            tag_names=tag_names,
            exclude_tag_names=exclude_tag_names,
            tag_group_id=tag_group_id,
            search_text=search_text,
            limit=limit,
            offset=offset,
            order_by=order_by,
            sort_order=sort_order,
            paginate=paginate,
            cache=cache,
            api_key=api_key,
        )


class Series(_api.API):
    """Get FRED series related to FRED tags.

    The class variable :data:`finagg.fred.api.tags.series` is an instance
    of this API implementation and is the most popular interface for calling
    this API.

    """

    url = "https://api.stlouisfed.org/fred/tags/series"

    @classmethod
    def get(
        cls,
        *,
        tag_names: None | str | list[str] = None,
        exclude_tag_names: None | str | list[str] = None,
        realtime_start: None | int | str = None,
        realtime_end: None | int | str = None,
        limit: None | int = 1000,
        offset: None | int = 0,
        order_by: None | str = None,
        sort_order: None | str = None,
        paginate: bool = False,
        cache: bool = True,
        api_key: None | str = None,
    ) -> pd.DataFrame:
        """Get the economic data series matching tags.

        See the related FRED API documentation at:

            https://fred.stlouisfed.org/docs/api/fred/tags_series.html

        Args:
            tag_names: Find series that match these tags.
            exclude_tag_names: Exclude series that match none of these tags.
            realtime_start: Start date for fetching results
                according to their publication date.
            realtime_end: End date for fetching results according
                to their publication date.
            limit: Maximum number of results to return.
            offset: Result start offset.
            order_by: Variable to order results by. Options include:

                - "series_id"
                - "title"
                - "units"
                - "frequency"
                - "seasonal_adjustment"
                - "realtime_start"
                - "realtime_end"
                - "last_updated"
                - "observation_start"
                - "observation_end"
                - "popularity"
                - "group_popularity"

            sort_order: Sort results in ascending ("asc") or
                descending ("desc") order.
            paginate: Whether to manage `offset` automatically, making multiple
                API calls until all results are returned.
            cache: Whether to cache the response from the API.
            api_key: Your FRED API key. Defaults to the ``FRED_API_KEY``
                environment variable.

        Returns:
            A dataframe containing series data for related tags.

        Examples:
            >>> finagg.fred.api.tags.series.get(tag_names="bea", limit=5)  # doctest: +SKIP
                            id realtime_start realtime_end                                             title ...
            0  A001RD3A086NBEA     2023-03-15   2023-03-15  Gross national product (implicit price deflator) ...
            1  A001RG3A086NBEA     2023-03-15   2023-03-15   Gross national product (chain-type price index) ...
            2  A001RI1A225NBEA     2023-03-15   2023-03-15   Gross National Product: Implicit Price Deflator ...
            3  A001RI1Q225SBEA     2023-03-15   2023-03-15   Gross National Product: Implicit Price Deflator ...
            4  A001RL1A225NBEA     2023-03-15   2023-03-15                       Real Gross National Product ...

        """
        return _api.maybe_paginate(
            "seriess",
            cls.url,
            tag_names=tag_names,
            exclude_tag_names=exclude_tag_names,
            realtime_start=realtime_start,
            realtime_end=realtime_end,
            limit=limit,
            offset=offset,
            order_by=order_by,
            sort_order=sort_order,
            paginate=paginate,
            cache=cache,
            api_key=api_key,
        )


class Tags(_api.API):
    """Get FRED tags.

    The module variable :data:`tags` is an instance of this API
    implementation and is the most popular interface for calling
    this API.

    """

    series = Series()
    """"tags/series" FRED API. Get the series for a FRED tag.
    The most popular way for accessing the :class:`Series` API.

    :meta hide-value:
    """

    url = "https://api.stlouisfed.org/fred/tags"

    @classmethod
    def get(
        cls,
        *,
        realtime_start: None | int | str = None,
        realtime_end: None | int | str = None,
        tag_names: None | str | list[str] = None,
        tag_group_id: None | str = None,
        search_text: None | str | list[str] = None,
        limit: None | int = 1000,
        offset: None | int = 0,
        order_by: None | str = None,
        sort_order: None | str = None,
        paginate: bool = False,
        cache: bool = True,
        api_key: None | str = None,
    ) -> pd.DataFrame:
        """Get the FRED tags for a series.

        See the related FRED API documentation at:

            https://fred.stlouisfed.org/docs/api/fred/tags.html

        Args:
            realtime_start: Start date for fetching results
                according to their publication date.
            realtime_end: End date for fetching results according
                to their publication date.
            tag_names: Filtering of tag names to include in the results.
            tag_group_id: A tag group ID to filter tags by. Options include:

                - "freq" = frequency
                - "gen" = general or concept
                - "geo" = geography
                - "geot" = geography type
                - "rls" = release
                - "seas" = seasonal adjustment
                - "src" = source
                - "cc" = citation and copyright

            search_text: The words to find matching tags with.
            limit: Maximum number of results to return.
            offset: Result start offset.
            order_by: Variable to order results by. Options include:

                - "series_count"
                - "popularity"
                - "created"
                - "name"
                - "group_id"

            sort_order: Sort results in ascending ("asc") or
                descending ("desc") order.
            paginate: Whether to manage `offset` automatically, making multiple
                API calls until all results are returned.
            cache: Whether to cache the response from the API.
            api_key: Your FRED API key. Defaults to the ``FRED_API_KEY``
                environment variable.

        Returns:
            A dataframe containing data for all FRED economic data tags.

        Examples:
            >>> finagg.fred.api.tags.get(tag_group_id="src", limit=5)  # doctest: +SKIP
                      name group_id                        notes                 created  popularity  series_count
            0       census      src                       Census  2012-02-27 10:18:19-06          79        237692
            1          bls      src   Bureau of Labor Statistics  2012-02-27 10:18:19-06          89        175376
            2  realtor.com      src                               2020-03-24 11:15:04-05          66         90632
            3          bea      src  Bureau of Economic Analysis  2012-02-27 10:18:19-06          78         78842
            4      frb stl      src                St. Louis Fed  2012-02-27 10:18:19-06          68         78442

        """
        return _api.maybe_paginate(
            "tags",
            cls.url,
            realtime_start=realtime_start,
            realtime_end=realtime_end,
            tag_names=tag_names,
            tag_group_id=tag_group_id,
            search_text=search_text,
            limit=limit,
            offset=offset,
            order_by=order_by,
            sort_order=sort_order,
            paginate=paginate,
            cache=cache,
            api_key=api_key,
        )



"""Features from FRED sources."""

from ._raw import Series
from ._refined import Economic

__all__ = [
    "economic",
    "series",
    "Economic",
    "Series",
]

economic = Economic()
"""The most popular way for accessing :class:`finagg.fred.feat.Economic`.

:meta hide-value:
"""

series = Series()
"""The most popular way for accessing :class:`finagg.fred.feat.Series`.

:meta hide-value:
"""



"""An implementation of the Federal Reserve Economic Data (FRED) API.

The FRED API provides methods for retrieving, searching, and describing
economic data from a variety of sources. The FRED API is one of the most
popular APIs in the financial industry.

A FRED API key is required to use this API. You can request a FRED API key
at the `FRED API webpage`_. You can pass your FRED API key directly to the
implemented API getters, or you can set the ``FRED_API_KEY`` environment
variable to have the FRED API key be passed to the implemented API getters
for you.

Alternatively, running ``finagg fred install`` (or the broader
``finagg install``) will prompt you where to acquire a FRED KEY and will
automatically store it in an ``.env`` file in your current working directory.
The environment variables set in that ``.env`` file will be loaded into your
shell upon using ``finagg`` (whether that be through the Python interface or
through the CLI tools).

See the official `FRED API docs`_ for more info on the FRED API.

.. _`FRED API docs`: https://fred.stlouisfed.org/docs/api/fred/
.. _`FRED API webpage`: https://fred.stlouisfed.org/docs/api/api_key.html

"""

from ._api import API
from ._category import (
    Category,
    CategoryChildren,
    CategoryRelated,
    CategoryRelatedTags,
    CategorySeries,
    CategoryTags,
)
from ._release import (
    Release,
    ReleaseDates,
    ReleaseRelatedTags,
    Releases,
    ReleasesDates,
    ReleaseSeries,
    ReleaseSources,
    ReleaseTables,
    ReleaseTags,
)
from ._series import (
    Series,
    SeriesCategories,
    SeriesObservations,
    SeriesRelease,
    SeriesSearch,
    SeriesSearchRelatedTags,
    SeriesSearchTags,
    SeriesTags,
    SeriesUpdates,
    SeriesVintageDates,
    popular_series,
)
from ._source import Source, SourceReleases, Sources
from ._tags import RelatedTags, Tags

__all__ = [
    "category",
    "popular_series",
    "related_tags",
    "release",
    "releases",
    "series",
    "source",
    "sources",
    "tags",
    "API",
    "Category",
    "CategoryChildren",
    "CategoryRelated",
    "CategoryRelatedTags",
    "CategorySeries",
    "CategoryTags",
    "RelatedTags",
    "Release",
    "ReleaseDates",
    "ReleaseRelatedTags",
    "Releases",
    "ReleasesDates",
    "ReleaseSeries",
    "ReleaseSources",
    "ReleaseTables",
    "ReleaseTags",
    "Series",
    "SeriesCategories",
    "SeriesObservations",
    "SeriesRelease",
    "SeriesSearch",
    "SeriesSearchRelatedTags",
    "SeriesSearchTags",
    "SeriesTags",
    "SeriesUpdates",
    "SeriesVintageDates",
    "Source",
    "SourceReleases",
    "Sources",
    "Tags",
]

category = Category()
"""The most popular way for accessing the :class:`finagg.fred.api.Category`.

:meta hide-value:
"""

release = Release()
"""The most popular way for accessing :class:`finagg.fred.api.Release`.

:meta hide-value:
"""

releases = Releases()
"""The most popular way for accessing :class:`finagg.fred.api.Releases`.

:meta hide-value:
"""

series = Series()
"""The most popular way for accessing :class:`finagg.fred.api.Series`.

:meta hide-value:
"""

source = Source()
"""The most popular way for accessing :class:`finagg.fred.api.Source`.

:meta hide-value:
"""

sources = Sources()
"""The most popular way for accessing :class:`finagg.fred.api.Sources`.

:meta hide-value:
"""

tags = Tags()
"""The most popular way for accessing :class:`finagg.fred.api.Tags`.

:meta hide-value:
"""

related_tags = RelatedTags()
"""The most popular way for accessing :class:`finagg.fred.api.RelatedTags`.

:meta hide-value:
"""



"""FRED CLI and tools."""

import logging
import os
from typing import Literal

import click

from .. import utils
from . import api as _api
from . import feat as _feat

logging.basicConfig(
    format="%(asctime)s | %(levelname)s | %(message)s", level=logging.INFO
)
logger = logging.getLogger(__name__)


@click.group(help="Federal Reserve Economic Data (FRED) tools.")
def entry_point() -> None:
    ...


@entry_point.command(
    help=(
        "Set the FRED API key, drop and recreate tables, "
        "and install the recommended tables into the SQL database."
    ),
)
@click.option(
    "--raw",
    type=click.Choice(["series"]),
    multiple=True,
    help=(
        "Raw tables to install. `series` indicates raw FRED economic series "
        "observations/measurements (e.g., consumer price index, gross domestic "
        "product, etc.). At least `series` must be specified to enable "
        "installing refined data using the `refined` flag."
    ),
)
@click.option(
    "--refined",
    type=click.Choice(["economic"]),
    multiple=True,
    help=(
        "Refined tables to install. This requires raw data to be "
        "installed beforehand using the `--raw` flag or for the "
        "`--raw` flag to be set when this option is provided."
    ),
)
@click.option(
    "--all",
    "-a",
    "all_",
    is_flag=True,
    default=False,
    help="Whether to install all defined tables (including all refined tables).",
)
@click.option(
    "--series",
    "-sid",
    multiple=True,
    help=(
        "FRED economic series whose data is attempted to be downloaded and "
        "inserted into the SQL tables. Multiple series can be specified by "
        "providing multiple `series` options, by separating series with a "
        "comma (e.g., `GDP,FEDFUNDS`), or by providing IDs in a CSV file by "
        "specifying a file path (e.g., `fred_series.txt`). The CSV file "
        "can be formatted such that there's one string per line or multiple "
        "strings per line (delimited by a comma). The strings specified "
        "by this option are combined with the strings specified by the "
        "`series-set` option."
    ),
)
@click.option(
    "--series-set",
    "-ss",
    "series_set",
    type=click.Choice(["economic"]),
    default=None,
    help=(
        "Set of FRED economic series whose data is attempted to be downloaded "
        "and inserted into the SQL tables. 'economic' indicates the recommended "
        "and most popular series (e.g., consumer price index, gross domestic "
        "product, etc.)."
    ),
)
@click.option(
    "--recreate-tables",
    "-r",
    is_flag=True,
    default=False,
    help=(
        "Whether to reset the tables associated with the install options by "
        "dropping and recreating them."
    ),
)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    default=False,
    help="Sets the log level to DEBUG to show installation errors for each series.",
)
def install(
    raw: list[Literal["series"]] = [],
    refined: list[Literal["economic"]] = [],
    all_: bool = False,
    series: list[str] = [],
    series_set: None | Literal["economic"] = None,
    recreate_tables: bool = False,
    verbose: bool = False,
) -> int:
    if verbose:
        logging.getLogger(__package__).setLevel(logging.DEBUG)

    if "FRED_API_KEY" not in os.environ:
        api_key = input(
            "Enter your FRED API key below.\n\n"
            "You can request a FRED API key at\n"
            "https://fred.stlouisfed.org/docs/api/api_key.html.\n\n"
            "FRED API key (leave blank and hit ENTER to skip): "
        ).strip()
        if not api_key:
            logger.warning(
                "An empty FRED API key was given. Skipping finagg.fred installation."
            )
            return 0
        p = utils.setenv("FRED_API_KEY", api_key)
        logger.info(f"FRED API key written to {p}")
    else:
        logger.info("FRED API key found in the environment")

    total_rows = 0
    all_raw = set()
    if all_:
        all_raw = {"series"}
    elif raw:
        all_raw = set(raw)

    all_series = utils.expand_csv(series)
    if all_raw:
        match series_set:
            case "economic":
                all_series |= set(_api.popular_series)

        if not all_series:
            logger.info(
                f"Skipping {__package__} installation because no series were "
                "provided (by the `series` option or by the `series-set` option)"
            )
            return total_rows

    if "series" in all_raw:
        total_rows += _feat.series.install(all_series, recreate_tables=recreate_tables)

    all_refined = set()
    if all_:
        all_refined = {"economic"}
    elif refined:
        all_refined = set(refined)

    if "economic" in all_refined:
        total_rows += _feat.economic.install(recreate_tables=recreate_tables)

    if all_ or all_refined or all_raw:
        if total_rows:
            logger.info(f"{total_rows} total rows inserted for {__package__}")
        else:
            logger.warning(
                f"No rows were inserted for {__package__}. This could be an error if"
                " installations were skipped. Set the verbose flag with the"
                " `--verbose/-v` option to enable debug logging."
            )
    else:
        logger.info(
            f"Skipping {__package__} installation because no installation "
            "options are provided"
        )
    return total_rows



"""An implementation of the Securities and Exchange Commission's (SEC) EDGAR API.

The SEC EDGAR API provides methods for retrieving XBRL data (e.g., earnings
per share) from financial statements and SEC filing submission histories
(e.g., 10-Q/10-K forms). The SEC EDGAR API is one of the few APIs that
provides historical and current company financial data for free.

An SEC EDGAR API user agent declaration is required to use this API.
The user agent should be of format ``FIRST_NAME LAST_NAME E_MAIL``. You
can pass your user agent directly to the implemented API getters, or you
can set the ``SEC_API_USER_AGENT`` environment variable to have the user agent
be passed to the implemented API getters for you.

Alternatively, running ``finagg sec install`` (or the broader
``finagg install``) will prompt you to enter an SEC EDGAR API user agent
and will automatically store it in an ``.env`` file in your current working
directory. The environment variables set in that ``.env`` file will be loaded
into your shell upon using ``finagg`` (whether that be through the Python
interface or through the CLI tools).

See the official `SEC EDGAR API docs`_ for more info on the SEC API.

.. _`SEC EDGAR API docs`: https://www.sec.gov/edgar/sec-api-documentation

"""

import logging
import os
from abc import ABC, abstractmethod
from datetime import datetime, timedelta
from functools import cache
from typing import Any, ClassVar, TypedDict
from zipfile import ZipFile

import pandas as pd
import requests
import requests_cache
from tqdm import tqdm

from .. import backend, ratelimit, utils

logging.basicConfig(
    format="%(asctime)s | %(levelname)s | %(message)s", level=logging.INFO
)
logger = logging.getLogger(__name__)

if backend.disable_http_cache:
    session = requests.Session()
else:
    session = requests_cache.CachedSession(
        str(backend.http_cache_path),
        expire_after=timedelta(weeks=1),
    )


class Concept(TypedDict):
    """A collection of XBRL tag (data field), SEC EDGAR taxonomy that
    tag is in, and units for that tag.

    """

    #: Valid tag within the given `taxonomy`.
    tag: str

    #: Valid SEC EDGAR taxonomy.
    #: See https://www.sec.gov/info/edgar/edgartaxonomies.shtml for taxonomies.
    taxonomy: str

    #: Currency the concept is in.
    units: str


class Frame(Concept):
    """A frame is just a :class:`Concept` with a flag indicating if it supports
    "instantaneous" frame data.

    """

    #: Whether to retrieve "instant" calendrical results.
    instant: bool


def _frame_to_concept(frame: Frame, /) -> Concept:
    """Helper for converting from a frame to a concept."""
    return {
        "tag": frame["tag"],
        "taxonomy": frame["taxonomy"],
        "units": "/".join(frame["units"].split("-per-")),
    }


class SubmissionsResult(TypedDict):
    #: Company metadata.
    metadata: dict[str, Any]

    #: Most recent company filings.
    filings: pd.DataFrame


class API(ABC):
    """Abstract SEC EDGAR API."""

    #: Request API URL.
    url: ClassVar[str]

    @classmethod
    @abstractmethod
    def get(cls, *args: Any, **kwargs: Any) -> pd.DataFrame | SubmissionsResult:
        """Main API getter method."""


class CompanyConcept(API):
    """Get all XBRL disclosures for a single company and concept (a taxonomy
    and tag) in a single dataframe.

    The module variable :data:`finagg.sec.api.company_concept` is an instance of
    this API implementation and is the most popular interface for querying this
    API.

    .. seealso::
        :data:`finagg.sec.api.popular_concepts`: For a list of popular company
            concepts that can be used with this method. The concepts described
            within this member are the most widely available concepts for
            SEC filers.

    """

    url = "https://data.sec.gov/api/xbrl/companyconcept/CIK{cik}/{taxonomy}/{tag}.json"

    @classmethod
    def get(
        cls,
        tag: str,
        /,
        *,
        cik: None | str = None,
        ticker: None | str = None,
        taxonomy: str = "us-gaap",
        units: str = "USD",
        cache: bool = True,
        user_agent: None | str = None,
    ) -> pd.DataFrame:
        """Return all XBRL disclosures for a single company (CIK)
        and concept (a taxonomy and tag) in a single dataframe.

        Args:
            tag: Valid tag within the given ``taxonomy``.
            cik: Company SEC CIK. Mutually exclusive with ``ticker``.
            ticker: Company ticker. Mutually exclusive with ``cik``.
            taxonomy: Valid SEC EDGAR taxonomy.
                See https://www.sec.gov/info/edgar/edgartaxonomies.shtml for taxonomies.
            units: Currency to view results in.
            cache: Whether to cache the response from the API.
            user_agent: Self-declared bot header. Defaults to the value
                found in the ``SEC_API_USER_AGENT`` environment variable.

        Returns:
            Dataframe with normalized column names.

        Raises:
            `ValueError`: If both a ``cik`` and ``ticker`` are provided or
                neither are provided.

        Examples:
            >>> finagg.sec.api.company_concept.get(
            ...     "EarningsPerShareBasic",
            ...     ticker="AAPL",
            ...     taxonomy="us-gaap",
            ...     units="USD/shares",
            ... ).head(5)  # doctest: +SKIP
                    start         end  value                  accn    fy  fp ...
            0  2006-10-01  2007-09-29   4.04  0001193125-09-214859  2009  FY ...
            1  2006-10-01  2007-09-29   4.04  0001193125-10-012091  2009  FY ...
            2  2007-09-30  2008-06-28   4.20  0001193125-09-153165  2009  Q3 ...
            3  2008-03-30  2008-06-28   1.21  0001193125-09-153165  2009  Q3 ...
            4  2007-09-30  2008-09-27   5.48  0001193125-09-214859  2009  FY ...

        """
        if bool(cik) == bool(ticker):
            raise ValueError("Must provide a `cik` or a `ticker`.")

        if ticker:
            cik = str(get_cik(ticker))

        cik = str(cik).zfill(10)
        url = cls.url.format(cik=cik, taxonomy=taxonomy, tag=tag)
        response = _get(url, cache=cache, user_agent=user_agent)
        content = response.json()
        units = content.pop("units")
        results_list = []
        for unit, data in units.items():  # type: ignore
            df = pd.DataFrame(data)
            df["units"] = unit
            results_list.append(df)
        results = pd.concat(results_list)
        for k, v in content.items():
            results[k] = v
        results["cik"] = cik
        return results.rename(columns={"entityName": "entity", "val": "value"})

    @classmethod
    def get_multiple_original(
        cls,
        concepts: list[Concept],
        *,
        cik: None | str = None,
        ticker: None | str = None,
        form: None | str = None,
        start: None | str = None,
        end: None | str = None,
        cache: bool = True,
        user_agent: None | str = None,
    ) -> pd.DataFrame:
        """Return original (not amended) XBRL disclosures for a single company
        using a set of company concepts.

        Args:
            concepts: Company concepts to retrieve.
            cik: Company SEC CIK. Mutually exclusive with ``ticker``.
            ticker: Company ticker. Mutually exclusive with ``cik``.
            form: SEC filing form type to include. ``"10-Q"`` is for quarterly
                filing forms and ``"10-K"`` is for annual filing forms. Ignored
                if left ``None``.
            start: The start date of the observation period. Defaults to the
                first recorded date.
            end: The end date of the observation period. Defaults to the
                last recorded date.
            cache: Whether to cache the response from the API.
            user_agent: Self-declared bot header. Defaults to the value
                found in the ``SEC_API_USER_AGENT`` environment variable.

        Returns:
            Pivoted dataframe with a tag per column.

        Examples:
            >>> finagg.sec.api.company_concept.get_multiple_original(
            ...     finagg.sec.api.popular_concepts,
            ...     ticker="AAPL",
            ... ).head(5)  # doctest: +SKIP
                fy  fp     tag         end         value  ...
            0  2009  Q3  Assets  2009-06-27  4.814000e+10 ...
            1  2010  Q1  Assets  2009-12-26  5.392600e+10 ...
            2  2010  Q2  Assets  2010-03-27  5.705700e+10 ...
            3  2010  Q3  Assets  2010-06-26  6.472500e+10 ...
            4  2011  Q1  Assets  2010-12-25  8.674200e+10 ...

        """
        start = start or "1776-07-04"
        end = end or utils.today
        dfs = []
        for concept in concepts:
            tag = concept["tag"]
            taxonomy = concept["taxonomy"]
            units = concept["units"]
            df = cls.get(
                tag,
                cik=cik,
                ticker=ticker,
                taxonomy=taxonomy,
                units=units,
                cache=cache,
                user_agent=user_agent,
            )
            df = filter_original_filings(df, form=form, units=units)
            df = df[(df["filed"] >= start) & (df["filed"] <= end)]
            dfs.append(df)
        df = pd.concat(dfs)
        return df


class CompanyFacts(API):
    """Get all XBRL disclosures for a single company.

    The module variable :data:`finagg.sec.api.company_facts` is an instance of
    this API implementation and is the most popular interface for querying this
    API.

    """

    url = "https://data.sec.gov/api/xbrl/companyfacts/CIK{cik}.json"

    @classmethod
    def download_zip(
        cls,
        /,
        *,
        chunk_size: int = 1_000_000,
        user_agent: None | str = None,
    ) -> ZipFile:
        """Download all XBRL disclosures for all companies in a zip file.

        A progress bar displays the download progress as the download could
        take a long time because the zip file is more than 1GB.
        The zip file is updated by the SEC at approximately 3am ET nightly.

        Args:
            chunk_size: Chunk size to stream and write the zip file with.
            user_agent: Self-declared bot header. Defaults to the value
                found in the ``SEC_API_USER_AGENT`` environment variable.

        Returns:
            A zip file object representing the downloaded zip file.

        """
        response = _get(
            "https://www.sec.gov/Archives/edgar/daily-index/xbrl/companyfacts.zip",
            cache=False,
            stream=True,
            user_agent=user_agent,
        )
        dst = backend.root_path / "findata" / "companyfacts.zip"
        pb = tqdm(
            total=int(response.headers.get("content-length", 0)),
            desc=f"Downloading companyfacts.zip to {dst}",
            position=0,
            leave=True,
            unit="iB",
            unit_scale=True,
        )
        with open(dst, "wb") as f:
            for chunk in response.iter_content(chunk_size):
                pb.update(len(chunk))
                f.write(chunk)
        pb.close()
        return ZipFile(dst)

    @classmethod
    def get(
        cls,
        /,
        *,
        cik: None | str = None,
        ticker: None | str = None,
        cache: bool = True,
        user_agent: None | str = None,
    ) -> pd.DataFrame:
        """Return all XBRL disclosures for a single company (CIK).

        Args:
            cik: Company SEC CIK. Mutually exclusive with `ticker`.
            ticker: Company ticker. Mutually exclusive with `cik`.
            cache: Whether to cache the response from the API.
            user_agent: Self-declared bot header. Defaults to the value
                found in the ``SEC_API_USER_AGENT`` environment variable.

        Returns:
            Dataframe with normalized column names.

        Raises:
            `ValueError`: If both a ``cik`` and ``ticker`` are provided or
                neither are provided.

        Examples:
            >>> finagg.sec.api.company_facts.get(ticker="AAPL").head(5)  # doctest: +SKIP
                      end       value                  accn    fy  fp    form ...
            0  2009-06-27  8.9582e+08  0001193125-09-153165  2009  Q3    10-Q ...
            1  2009-10-16  9.0068e+08  0001193125-09-214859  2009  FY    10-K ...
            2  2009-10-16  9.0068e+08  0001193125-10-012091  2009  FY  10-K/A ...
            3  2010-01-15  9.0679e+08  0001193125-10-012085  2010  Q1    10-Q ...
            4  2010-04-09  9.0994e+08  0001193125-10-088957  2010  Q2    10-Q ...

        """
        if bool(cik) == bool(ticker):
            raise ValueError("Must provide a `cik` or a `ticker`.")

        if ticker:
            cik = str(get_cik(ticker))

        cik = str(cik).zfill(10)
        url = cls.url.format(cik=cik)
        response = _get(url, cache=cache, user_agent=user_agent)
        content = response.json()
        df = _parse_company_facts(content)
        df["cik"] = cik
        return df


class Exchanges(API):
    """SEC-registered ticker info with exchange data.

    This API implementation is very similar to :class:`Tickers`. It returns
    all the same data in addition to the exchanges that each company is on.

    The module variable :data:`finagg.sec.api.exchanges` is an instance of this
    API implementation and is the most popular interface for querying this
    API.

    """

    url = "https://www.sec.gov/files/company_tickers_exchange.json"

    @classmethod
    def get(cls, *, cache: bool = True, user_agent: None | str = None) -> pd.DataFrame:
        """Get a dataframe containing all SEC-registered ticker
        info.

        Args:
            cache: Whether to cache the response from the API.
            user_agent: Self-declared SEC bot header. Defaults to the value
                found in the ``SEC_API_USER_AGENT`` environment variable.

        Returns:
            A dataframe containing company names, their SEC CIKs, and their
            ticker symbols.

        Examples:
            >>> finagg.sec.api.exchanges.get().head(5)  # doctest: +SKIP
                   cik            name ticker exchange
            0   320193      Apple Inc.   AAPL   Nasdaq
            1   789019  MICROSOFT CORP   MSFT   Nasdaq
            2  1652044   Alphabet Inc.  GOOGL   Nasdaq
            3  1018724  AMAZON COM INC   AMZN   Nasdaq
            4  1045810     NVIDIA CORP   NVDA   Nasdaq

        """
        response = _get(cls.url, cache=cache, user_agent=user_agent)
        content: dict[str, list[str]] = response.json()
        df = pd.DataFrame(content["data"], columns=content["fields"])
        return df.rename(columns={"cik_str": "cik"})


class Frames(API):
    """Get all company filings for one particular fact that most closely
    matches the requested calendrical period.

    The module variable :data:`finagg.sec.api.frames` is an instance of
    this API implementation and is the most popular interface for querying this
    API.

    .. seealso::
        :data:`popular_frames`: For a list of popular company frames
            that can be used with this method. The frames described
            within this member are the most widely available frames for
            SEC filers.

    """

    url = (
        "https://data.sec.gov/api/xbrl"
        "/frames"
        "/{taxonomy}/{tag}/{units}/CY{year}{quarter}.json"
    )

    @classmethod
    def get(
        cls,
        tag: str,
        year: int | str,
        /,
        quarter: None | int | str = None,
        *,
        instant: bool = True,
        taxonomy: str = "us-gaap",
        units: str = "USD",
        cache: bool = True,
        user_agent: None | str = None,
    ) -> pd.DataFrame:
        """Get one fact for each reporting entity that most closely fits
        the calendrical period requested.

        Args:
            tag: Valid tag within the given ``taxonomy``.
            year: Year to retrieve.
            quarter: Quarter to retrieve data for. Most data is only provided
                at a quarterly rate, so this should be provided for most cases.
            instant: Whether to get instantaneous data for the frame (data that
                most closely matches a frame's year and quarter without a
                time buffer). This flag should be enabled for most cases. See
                :data:`popular_frames` for which tags should be ``instant``.
            taxonomy: Valid SEC EDGAR taxonomy.
                See https://www.sec.gov/info/edgar/edgartaxonomies for taxonomies.
            units: Units to view results in.
            cache: Whether to cache the response from the API.
            user_agent: Self-declared bot header. Defaults to the value
                found in the ``SEC_API_USER_AGENT`` environment variable.

        Returns:
            Dataframe with slightly improved column names.

        Examples:
            >>> finagg.sec.api.frames.get(
            ...     "EarningsPerShareBasic",
            ...     2020,
            ...     quarter=3,
            ...     instant=False,
            ...     taxonomy="us-gaap",
            ...     units="USD-per-shares",
            ... ).head(5)  # doctest: +SKIP
                               accn   cik                          entity    loc ...
            0  0001104659-21-118843  1750                       AAR CORP.  US-IL ...
            1  0001104659-21-133629  1800             ABBOTT LABORATORIES  US-IL ...
            2  0001264931-20-000235  1961                     WORLDS INC.  US-MA ...
            3  0001564590-21-055818  2098                ACME UNITED CORP  US-CT ...
            4  0000002178-22-000033  2178  ADAMS RESOURCES & ENERGY, INC.  US-TX ...

        """
        if quarter:
            quarter = f"Q{quarter}"
            if instant:
                quarter = f"{quarter}I"
        else:
            quarter = ""
        units = "-per-".join(units.split("/"))
        url = cls.url.format(
            taxonomy=taxonomy, tag=tag, units=units, year=year, quarter=quarter
        )
        response = _get(url, cache=cache, user_agent=user_agent)
        content = response.json()
        data = content.pop("data")
        df = pd.DataFrame(data)
        for k, v in content.items():
            df[k] = v
        return df.rename(
            columns={
                "ccp": "frame",
                "entityName": "entity",
                "uom": "units",
                "val": "value",
            }
        )


class Submissions(API):
    """Get a company's metadata and all its recent SEC filings.

    Not all the metadata typically found with the submissions API is supported
    by this API. Only common company metadata (e.g., company name, industry
    code, fiscal year end date, etc.) is returned by this implementation.

    The module variable :data:`finagg.sec.api.submissions` is an instance of
    this API implementation and is the most popular interface for querying this
    API.

    """

    url = "https://data.sec.gov/submissions/CIK{cik}.json"

    @classmethod
    def download_zip(
        cls,
        /,
        *,
        chunk_size: int = 1_000_000,
        user_agent: None | str = None,
    ) -> ZipFile:
        """Download all SEC filings for all companies in a zip file.

        A progress bar displays the download progress as the download could
        take a long time because the zip file is more than 1GB.
        The zip file is updated by the SEC at approximately 3am ET nightly.

        Args:
            chunk_size: Chunk size to stream and write the zip file with.
            user_agent: Self-declared bot header. Defaults to the value
                found in the ``SEC_API_USER_AGENT`` environment variable.

        Returns:
            A zip file object representing the downloaded zip file.

        """
        response = _get(
            "https://www.sec.gov/Archives/edgar/daily-index/bulkdata/submissions.zip",
            cache=False,
            stream=True,
            user_agent=user_agent,
        )
        dst = backend.root_path / "findata" / "submissions.zip"
        pb = tqdm(
            total=int(response.headers.get("content-length", 0)),
            desc=f"Downloading submissions.zip to {dst}",
            position=0,
            leave=True,
            unit="iB",
            unit_scale=True,
        )
        with open(dst, "wb") as f:
            for chunk in response.iter_content(chunk_size):
                pb.update(len(chunk))
                f.write(chunk)
        pb.close()
        return ZipFile(dst)

    @classmethod
    def get(
        cls,
        /,
        *,
        cik: None | str = None,
        ticker: None | str = None,
        cache: bool = True,
        user_agent: None | str = None,
    ) -> SubmissionsResult:
        """Return a company's metadata and all its recent SEC filings.

        Args:
            cik: Company SEC CIK. Mutually exclusive with ``ticker``.
            ticker: Company ticker. Mutually exclusive with ``cik``.
            cache: Whether to cache the response from the API.
            user_agent: Self-declared SEC bot header. Defaults to the value
                found in the ``SEC_API_USER_AGENT`` environment variable.

        Returns:
            Company metadata and a dataframe with recent SEC filings.

        Raises:
            `ValueError`: If both a ``cik`` and ``ticker`` are provided or neither
                are provided.

        Examples:
            >>> out = finagg.sec.api.submissions.get(ticker="AAPL")
            >>> out["metadata"]  # doctest: +ELLIPSIS
            {'cik': '0000320193', 'entityType': 'operating', 'sic': '3571', ...}

        """
        if bool(cik) == bool(ticker):
            raise ValueError("Must provide a `cik` or a `ticker`.")

        if ticker:
            cik = str(get_cik(ticker))

        if cik:
            ticker = str(get_ticker(cik))

        cik = str(cik).zfill(10)
        url = cls.url.format(cik=cik)
        response = _get(url, cache=cache, user_agent=user_agent)
        content = response.json()
        recent_filings = content.pop("filings")["recent"]
        df = pd.DataFrame(recent_filings)
        df.columns = map(utils.snake_case, df.columns)  # type: ignore
        df.rename(columns={"accession_number": "accn"})
        df["cik"] = cik
        metadata = _parse_submission_metadata(content)
        metadata["cik"] = cik
        metadata["ticker"] = str(ticker)
        return {"metadata": metadata, "filings": df}


class Tickers(API):
    """SEC-registered ticker info.

    This is a broader method in comparison to :meth:`get_ticker_set`.
    :meth:`get_ticker_set` will get all the tickers that have popular
    fundamentals data available through the SEC EDGAR API for the
    previous year, while this method will get all tickers and CIKs
    that have ever had data available via the SEC EDGAR API.

    The module variable :data:`finagg.sec.api.tickers` is an instance of this
    API implementation and is the most popular interface for querying this
    API.

    """

    url = "https://www.sec.gov/files/company_tickers.json"

    @classmethod
    def get(cls, *, cache: bool = True, user_agent: None | str = None) -> pd.DataFrame:
        """Get a dataframe containing all SEC-registered ticker
        info.

        Args:
            cache: Whether to cache the response from the API.
            user_agent: Self-declared SEC bot header. Defaults to the value
                found in the ``SEC_API_USER_AGENT`` environment variable.

        Returns:
            A dataframe containing company names, their SEC CIKs, and their
            ticker symbols.

        Examples:
            >>> finagg.sec.api.tickers.get().head(5)  # doctest: +SKIP
                   cik ticker                   title
            0   320193   AAPL              Apple Inc.
            1   789019   MSFT          MICROSOFT CORP
            2  1652044  GOOGL           Alphabet Inc.
            3  1018724   AMZN          AMAZON COM INC
            4  1067983  BRK-B  BERKSHIRE HATHAWAY INC

        """
        response = _get(cls.url, cache=cache, user_agent=user_agent)
        content: dict[str, dict[str, str]] = response.json()
        df = pd.DataFrame([items for _, items in content.items()])
        return df.rename(columns={"cik_str": "cik"})


# Mapping of SEC CIK strings to (uppercase) tickers.
_cik_to_tickers: dict[str, str] = {}

# Mapping of (uppercase) tickers to SEC CIK strings.
_tickers_to_cik: dict[str, str] = {}

company_concept = CompanyConcept()
"""The most popular way for accessing the :class:`CompanyConcept` API
implementation.

:meta hide-value:
"""

company_facts = CompanyFacts()
"""The most popular way for accessing the :class:`CompanyFacts` API
implementation.

:meta hide-value:
"""

exchanges = Exchanges()
"""The most popular way for accessing the :class:`Exchanges` API implementation.

:meta hide-value:
"""

frames = Frames()
"""The most popular way for accessing the :class:`Frames` API implementation.

:meta hide-value:
"""

submissions = Submissions()
"""The most popular way for accessing the :class:`Submissions` API
implementation.

:meta hide-value:
"""

tickers = Tickers()
"""The most popular way for accessing the :class:`Tickers` API implementation.

:meta hide-value:
"""

popular_frames: list[Frame] = [
    {
        "tag": "Assets",
        "taxonomy": "us-gaap",
        "units": "USD",
        "instant": True,
    },
    {
        "tag": "AssetsCurrent",
        "taxonomy": "us-gaap",
        "units": "USD",
        "instant": True,
    },
    {
        "tag": "CommonStockSharesOutstanding",
        "taxonomy": "us-gaap",
        "units": "shares",
        "instant": True,
    },
    {
        "tag": "EarningsPerShareBasic",
        "taxonomy": "us-gaap",
        "units": "USD-per-shares",
        "instant": False,
    },
    {"tag": "InventoryNet", "taxonomy": "us-gaap", "units": "USD", "instant": True},
    {
        "tag": "Liabilities",
        "taxonomy": "us-gaap",
        "units": "USD",
        "instant": True,
    },
    {
        "tag": "LiabilitiesCurrent",
        "taxonomy": "us-gaap",
        "units": "USD",
        "instant": True,
    },
    {
        "tag": "NetIncomeLoss",
        "taxonomy": "us-gaap",
        "units": "USD",
        "instant": True,
    },
    {
        "tag": "StockholdersEquity",
        "taxonomy": "us-gaap",
        "units": "USD",
        "instant": True,
    },
]
"""Company frames that have high availability and are relatively popular
for fundamental analysis. Includes things like earnings per share, current
assets, etc.. Frames are in valid format for usage with the :class:`Frames`
API implementation.

:meta hide-value:
"""

popular_concepts: list[Concept] = [_frame_to_concept(frame) for frame in popular_frames]
"""Company concepts that have high availability and are relatively popular
for fundamental analysis. Includes things like earnings per share, current
assets, etc.. Concepts contain all the parameters required for usage with the
:class:`CompanyConcept` API implementation.

:meta hide-value:
"""


@ratelimit.guard([ratelimit.RequestLimit(9, timedelta(seconds=1))])
def _get(
    url: str,
    /,
    *,
    stream: bool = False,
    cache: bool = True,
    user_agent: None | str = None,
) -> requests.Response:
    """SEC EDGAR API request helper.

    Args:
        url: Complete SEC EDGAR API URL to request from.
        stream: Whether to stream the response content (useful for file
            downloads).
        cache: Whether to cache the response from the given URL.
        user_agent: Self-declared SEC bot header. Defaults to the value
            found in the ``SEC_API_USER_AGENT`` environment variable.

    Returns:
        Successful responses.

    Raises:
        `RuntimeError`: If a user agent is not provided or found in the environment.

    """
    user_agent = user_agent or os.environ.get("SEC_API_USER_AGENT", None)
    if not user_agent:
        raise RuntimeError(
            "No SEC API user agent declaration found. "
            "Pass your user agent declaration to the API directly, or "
            "set the `SEC_API_USER_AGENT` environment variable."
        )
    headers = {"User-Agent": user_agent}
    if cache:
        response = session.get(url, headers=headers, stream=stream)
    else:
        response = requests.get(url, headers=headers, stream=stream)
    response.raise_for_status()
    return response


def compute_financial_ratios(df: pd.DataFrame, /) -> pd.DataFrame:
    """Compute financial ratios in-place for a dataframe using XBRL financial
    tags.

    Args:
        df: Dataframe with common XBRL financial tags (e.g., "Assets",
            "Liabilities", etc.).

    Returns:
        ``df`` but with additional columns that're normalized financial ratios.

    """
    df["AssetCoverageRatio"] = (df["Assets"] - df["LiabilitiesCurrent"]) / df[
        "Liabilities"
    ]
    df["BookRatio"] = (df["Assets"] - df["Liabilities"]) / df[
        "CommonStockSharesOutstanding"
    ]
    df["DebtEquityRatio"] = df["Liabilities"] / df["StockholdersEquity"]
    df["QuickRatio"] = (df["AssetsCurrent"] - df["InventoryNet"]) / df[
        "LiabilitiesCurrent"
    ]
    df["ReturnOnAssets"] = df["NetIncomeLoss"] / df["Assets"]
    df["ReturnOnEquity"] = df["NetIncomeLoss"] / df["StockholdersEquity"]
    df["WorkingCapitalRatio"] = df["AssetsCurrent"] / df["LiabilitiesCurrent"]
    return df


def filter_original_filings(
    df: pd.DataFrame, /, *, form: None | str = None, units: None | str = None
) -> pd.DataFrame:
    """Get all original filing rows as determined by the filing date and
    tag for a period.

    Args:
        df: Dataframe without mixed original and amendment rows.
        form: Only keep rows with form type ``form``. Most popular choices
            include ``"10-K"`` for annual and ``"10-Q"`` for quarterly.
            This parameter is ignored if left ``None``.
        units: Only keep rows with units ``units`` if not ``None``.

    Returns:
        Dataframe with rows that only correspond to original filings.

    Examples:
        Only get a company's original quarterly earnings-per-share filings.

        >>> df = finagg.sec.api.company_concept.get(
        ...     "EarningsPerShareBasic",
        ...     ticker="AAPL",
        ...     taxonomy="us-gaap",
        ...     units="USD/shares",
        ... )
        >>> finagg.sec.api.filter_original_filings(
        ...     df,
        ...     form="10-Q",
        ...     units="USD/shares"
        ... ).head(5)  # doctest: +SKIP
             fy  fp  ...
        0  2009  Q3  ...
        1  2010  Q1  ...
        2  2010  Q2  ...
        3  2010  Q3  ...
        4  2011  Q1  ...

    """
    end = pd.DatetimeIndex(df["end"])
    filed = pd.DatetimeIndex(df["filed"])
    filing_delay = filed - end
    # Make sure filings occurs within 90 days of the reporting end date.
    # Helps ensure each filing is the first filing and not an amendment.
    mask = filing_delay.days <= 90
    if form:
        mask &= df["form"] == form
        # Not all filings contain a start date, but it can be helpful to
        # use the start date to ensure the filing corresponds to the time
        # period we care about.
        if "start" in df:
            start: pd.DatetimeIndex = pd.DatetimeIndex(df["start"])
            start_to_end = end - start
        match form:
            case "10-K":
                mask &= df["fp"] == "FY"
                # Make sure the reporting frame is close to a year.
                if "start" in df:
                    mask &= start.isna() | ((350 <= start_to_end.days) & (start_to_end.days <= 380))  # type: ignore[no-untyped-call]
            case "10-Q":
                mask &= df["fp"].str.startswith("Q")
                # Make sure the reporting frame is close to a quarter.
                if "start" in df:
                    mask &= start.isna() | ((75 <= start_to_end.days) & (start_to_end.days <= 105))  # type: ignore[no-untyped-call]
    if units:
        mask &= df["units"] == units
    df = df[mask]
    return (
        df.sort_values(["fy", "fp", "filed"])
        .groupby(["fy", "fp", "tag"], as_index=False)
        .first()
    )


def get_cik(ticker: str, /, *, user_agent: None | str = None) -> str:
    """Return a company's SEC CIK from its ticker.

    Args:
        ticker: A company's ticker symbol.
        user_agent: Self-declared SEC bot header. Defaults to the value
            found in the ``SEC_API_USER_AGENT`` environment variable.

    Returns:
        The company's corresponding SEC CIK.

    Examples:
        Get Apple's SEC CIK from its ticker.

        >>> finagg.sec.api.get_cik("AAPL") == "0000320193"
        True

    """
    if not _tickers_to_cik:
        response = _get(Tickers.url, user_agent=user_agent)
        content: dict[str, dict[str, str]] = response.json()
        for _, items in content.items():
            normalized_cik = str(items["cik_str"]).zfill(10)
            _tickers_to_cik[items["ticker"]] = normalized_cik
            _cik_to_tickers[normalized_cik] = items["ticker"]
    return _tickers_to_cik[ticker.upper()]


def get_ticker(cik: str, /, *, user_agent: None | str = None) -> str:
    """Return a company's ticker from its SEC CIK.

    Args:
        cik: A company's 10-character SEC CIK.
        user_agent: Self-declared SEC bot header. Defaults to the value
            found in the ``SEC_API_USER_AGENT`` environment variable.

    Returns:
        The company's corresponding ticker.

    Examples:
        Get Apple's ticker from its SEC CIK.

        >>> finagg.sec.api.get_ticker("0000320193") == "AAPL"
        True

    """
    if not _cik_to_tickers:
        response = _get(Tickers.url, user_agent=user_agent)
        content: dict[str, dict[str, str]] = response.json()
        for _, items in content.items():
            normalized_cik = str(items["cik_str"]).zfill(10)
            _cik_to_tickers[normalized_cik] = items["ticker"]
            _tickers_to_cik[items["ticker"]] = normalized_cik
    cik = str(cik).zfill(10)
    return _cik_to_tickers[cik]


@cache
def get_ticker_set(*, user_agent: None | str = None) -> set[str]:
    """Get the set of tickers that published data for popular concepts
    during any of the quarters for the previous year.

    This effectively gets the set of tickers whose data is at least
    somewhat available through the SEC EDGAR API.

    Args:
        user_agent: Self-declared SEC bot header. Defaults to the value
            found in the ``SEC_API_USER_AGENT`` environment variable.

    Returns:
        Set of tickers whose data is at least somewhat available through
        the SEC EDGAR API.

    Examples:
        >>> tickers = finagg.sec.api.get_ticker_set()
        >>> "AAPL" in tickers
        True
        >>> "MSFT" in tickers
        True
        >>> "GOOG" in tickers
        True

    """
    year = datetime.now().year - 1
    tickers = set()

    for frame in popular_frames:
        tag = frame["tag"]
        instant = frame["instant"]
        taxonomy = frame["taxonomy"]
        units = frame["units"]
        for quarter in range(1, 4):
            df = frames.get(
                tag,
                year,
                quarter,
                instant=instant,
                taxonomy=taxonomy,
                units=units,
                user_agent=user_agent,
            )
            for cik in df["cik"]:
                try:
                    ticker = get_ticker(cik)
                except KeyError:
                    continue
                tickers.add(ticker)
    return tickers


def group_and_pivot_filings(
    df: pd.DataFrame, /, *, form: None | str = None
) -> pd.DataFrame:
    """Helper for grouping filings into a pivoted dataframe such that each
    tag has its own column.

    Tags are grouped according to fiscal year and fiscal period. The newest
    filing date is used when tags have different filing dates for the same
    fiscal year and fiscal period. Tags are forward-filled to fill gaps
    in filings.

    Args:
        df: "Melted" dataframe where each filing is a row.
        form: Type of form to group. ``"10-K"`` sets the index to the fiscal
            year and filing date whereas ``"10-Q"`` sets the index to the
            fiscal year, fiscal period, and filing date. If not provided,
            the form type is chosen by inspecting the first element of the
            ``"form"`` column.

    Returns:
        A pivoted dataframe where each column is a tag.

    Examples:
        Get all XBRL filings from a company, unintelligently select the first
        unique filings for each XBRL tag, and then group all filings into
        a pivoted table.

        >>> df = finagg.sec.api.company_facts.get(ticker="AAPL")
        >>> df = finagg.sec.api.filter_original_filings(df, form="10-K")
        >>> finagg.sec.api.group_and_pivot_filings(df, form="10-K")  # doctest: +SKIP
                         AccountsPayableCurrent  AccountsReceivableNetCurrent ...
        fy   filed                                                            ...
        2009 2009-10-27            5.520000e+09                  2.422000e+09 ...
        2010 2010-10-27            5.601000e+09                  3.361000e+09 ...
        2011 2011-10-26            1.201500e+10                  5.510000e+09 ...
        2012 2012-10-31            1.463200e+10                  5.369000e+09 ...
        2013 2013-10-30            2.117500e+10                  1.093000e+10 ...

    """
    if form is None:
        form = df.iloc[0]["form"]
    match form:
        case "10-K":
            df = df.drop(columns=["fp"]).set_index(["fy"]).sort_index()
            df["filed"] = df.groupby(["fy"])["filed"].max()
            df = df.reset_index().pivot(
                index=["fy", "filed"],
                columns="tag",
                values="value",
            )
        case "10-Q":
            df = df.set_index(["fy", "fp"]).sort_index()
            df["filed"] = df.groupby(["fy", "fp"])["filed"].max()
            df = df.reset_index().pivot(
                index=["fy", "fp", "filed"],
                columns="tag",
                values="value",
            )
    df.columns = df.columns.rename(None)
    return df


def _parse_company_facts(content: dict[str, Any], /) -> pd.DataFrame:
    """Helper for parsing company facts.

    This function is only defined to make parsing company
    facts easier and common between parsing from the REST
    API responses and the bulk zip file.

    Args:
        content: Company facts JSON data.

    Returns:
        A dataframe equivalent of the company facts data.

    """
    facts = content.pop("facts")
    results_list = []
    for taxonomy, tag_dict in facts.items():
        for tag, data in tag_dict.items():
            for col, rows in data["units"].items():
                df = pd.DataFrame(rows)
                df["taxonomy"] = taxonomy
                df["tag"] = tag
                df["label"] = data["label"]
                df["description"] = data["description"]
                df["units"] = col
                results_list.append(df)
    results = pd.concat(results_list)
    for k, v in content.items():
        results[k] = v
    return results.rename(
        columns={"entityName": "entity", "uom": "units", "val": "value"}
    )


def _parse_submission_metadata(content: dict[str, Any], /) -> dict[str, Any]:
    """Helper for parsing submission metadata.

    This function is only defined to make parsing submission
    metadata easier and common between parsing from the REST
    API responses and the bulk zip file.

    Args:
        content: Submissions JSON data.

    Returns:
        A dataframe equivalent of the submissions data.

    """
    metadata = {}
    for k, v in content.items():
        if isinstance(v, str):
            metadata[k] = utils.snake_case(v)
    if "exchanges" in content:
        metadata["exchanges"] = ",".join(content["exchanges"])
    return metadata



"""Fundamental features aggregated from several sources."""

import logging
import multiprocessing as mp
from datetime import datetime, timedelta
from typing import Literal

import numpy as np
import pandas as pd
import sqlalchemy as sa
from sqlalchemy.engine import Engine
from sqlalchemy.exc import NoResultFound

from .. import backend, sec, utils, yfinance
from . import sql

logging.basicConfig(
    format="%(asctime)s | %(levelname)s | %(message)s", level=logging.INFO
)
logger = logging.getLogger(__name__)


class IndustryFundamental:
    """Methods for gathering industry-averaged fundamental data.

    The class variable :attr:`finagg.fundam.feat.Fundamental.industry` is an
    instance of this feature set implementation and is the most popular
    interface for calling feature methods.

    Examples:
        You can aggregate this feature set using a ticker or an industry code
        directly.

        >>> df1 = finagg.fundam.feat.fundam.industry.from_refined(ticker="MSFT").head(5)
        >>> df2 = finagg.fundam.feat.fundam.industry.from_refined(code=73).head(5)
        >>> pd.testing.assert_frame_equal(df1, df2, rtol=1e-4)

    """

    @classmethod
    def from_refined(
        cls,
        /,
        *,
        ticker: None | str = None,
        code: None | str = None,
        level: Literal[2, 3, 4] = 2,
        start: None | str = None,
        end: None | str = None,
        engine: None | Engine = None,
    ) -> pd.DataFrame:
        """Get fundamental features from the feature store,
        aggregated for an entire industry.

        The industry can be chosen according to a company or
        by an industry code directly. If a company is provided,
        then the first ``level`` digits of the company's SIC code
        is used for the industry code.

        Args:
            ticker: Company ticker. Lookup the industry associated
                with this company. Mutually exclusive with ``code``.
            code: Industry SIC code to use for industry lookup.
                Mutually exclusive with ``ticker``.
            level: Industry level to aggregate features at.
                The industry used according to ``ticker`` or ``code``
                is subsampled according to this value. Options include:

                    - 2 = major group (e.g., furniture and fixtures)
                    - 3 = industry group (e.g., office furnitures)
                    - 4 = industry (e.g., wood office furniture)

            start: The start date of the observation period. Defaults to the
                first recorded date.
            end: The end date of the observation period. Defaults to the
                last recorded date.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            Fundamental data dataframe with each feature as a
            separate column. Sorted by date.

        Raises:
            `ValueError`: If neither a ``ticker`` nor ``code`` are provided.
            `NoResultFound`: If there are no rows for ``ticker`` or ``code``
                in the refined SQL table.

        Examples:
            >>> df = finagg.fundam.feat.fundam.industry.from_refined(ticker="AAPL").head(5)
            >>> df["mean"]  # doctest: +SKIP
            name        PriceBookRatio  PriceEarningsRatio
            date
            2010-01-29        1.213018           17.020706
            2010-02-01        1.166246           16.364503
            2010-02-02        1.213047           17.021558
            2010-02-03        1.222630           17.156233
            2010-02-04        1.213401           17.026466
            >>> df["std"]  # doctest: +SKIP
            name        PriceBookRatio  PriceEarningsRatio
            date
            2010-01-29        1.469641           20.667755
            2010-02-01        1.414144           19.887156
            2010-02-02        1.476001           20.756459
            2010-02-03        1.490459           20.959443
            2010-02-04        1.475500           20.749523

        """
        start = start or "1776-07-04"
        end = end or utils.today
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sec.sql.submissions.name):
            sec.sql.submissions.create(engine)
        if not sa.inspect(engine).has_table(sql.fundam.name):
            sql.fundam.create(engine)
        with engine.begin() as conn:
            if ticker:
                (sic,) = conn.execute(
                    sa.select(sec.sql.submissions.c.sic).where(
                        sec.sql.submissions.c.ticker == ticker
                    )
                ).one()
                code = str(sic)[:level]
            elif code:
                code = str(code)[:level]
            else:
                raise ValueError("Must provide a `ticker` or `code`.")

            df = pd.DataFrame(
                conn.execute(
                    sql.fundam.select()
                    .join(
                        sec.sql.submissions,
                        (sec.sql.submissions.c.ticker == sql.fundam.c.ticker)
                        & (sec.sql.submissions.c.sic.startswith(code)),
                    )
                    .where(sql.fundam.c.date >= start, sql.fundam.c.date <= end)
                )
            )
        if not len(df.index):
            raise NoResultFound(f"No industry fundamental rows found for {code}.")
        df = df.drop(columns=["ticker"])
        df = df.melt(["date"], var_name="name", value_name="value").set_index("date")
        return (
            df.groupby(["date", "name"])  # type: ignore[return-value]
            .agg(["mean", "std"])
            .reset_index()
            .pivot(index=["date"], columns="name")["value"]
            .sort_index()
            .dropna()
        )


class NormalizedFundamental:
    """Fundamental features from quarterly and daily data, normalized
    according to industry averages and standard deviations.

    The class variable :attr:`finagg.fundam.feat.Fundamental.normalized` is an
    instance of this feature set implementation and is the most popular
    interface for calling feature methods.

    Examples:
        It doesn't matter which data source you use to gather features.
        They all return equivalent dataframes.

        >>> df1 = finagg.fundam.feat.fundam.normalized.from_other_refined("AAPL").head(5)
        >>> df2 = finagg.fundam.feat.fundam.normalized.from_refined("AAPL").head(5)
        >>> pd.testing.assert_frame_equal(df1, df2, rtol=1e-4)

    """

    @classmethod
    def from_other_refined(
        cls,
        ticker: str,
        /,
        *,
        level: Literal[2, 3, 4] = 2,
        start: None | str = None,
        end: None | str = None,
        engine: None | Engine = None,
    ) -> pd.DataFrame:
        """Get features from other feature SQL tables.

        Args:
            ticker: Company ticker.
            level: Industry level to aggregate relative features at.
                The industry used according to ``ticker`` is subsampled
                according to this value. Options include:

                    - 2 = major group (e.g., furniture and fixtures)
                    - 3 = industry group (e.g., office furnitures)
                    - 4 = industry (e.g., wood office furniture)

            start: The start date of the observation period. Defaults to the
                first recorded date.
            end: The end date of the observation period. Defaults to the
                last recorded date.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            Relative fundamental data dataframe with each feature as a
            separate column. Sorted by filing date.

        Examples:
            >>> finagg.fundam.feat.fundam.normalized.from_other_refined("AAPL").head(5)  # doctest: +SKIP
                        NORM(PriceBookRatio)  NORM(PriceEarningsRatio)
            date
            2010-01-25             -0.706266                 -0.706279
            2010-01-26             -0.698805                 -0.698935
            2010-01-27             -0.700699                 -0.700799
            2010-01-28             -0.701446                 -0.701534
            2010-01-29             -0.704558                 -0.704598

        """
        start = start or "1776-07-04"
        end = end or utils.today
        engine = engine or backend.engine
        company_df = Fundamental.from_refined(
            ticker, start=start, end=end, engine=engine
        ).reset_index(["date"])
        date = company_df["date"]
        industry_df = IndustryFundamental.from_refined(
            ticker=ticker, level=level, start=start, end=end, engine=engine
        ).reset_index(["date"])
        company_df = (company_df - industry_df["mean"]) / industry_df["std"]
        company_df["date"] = date
        company_df = (
            company_df.reset_index()
            .set_index("date")
            .rename(lambda x: f"NORM({x})", axis=1)
        )
        company_df = (
            company_df.ffill()
            .reset_index()
            .dropna()
            .drop_duplicates("date")
            .set_index("date")
            .sort_index()
        )
        company_df = utils.resolve_col_order(sql.normalized_fundam, company_df)
        return company_df

    @classmethod
    def from_refined(
        cls,
        ticker: str,
        /,
        *,
        start: None | str = None,
        end: None | str = None,
        engine: None | Engine = None,
    ) -> pd.DataFrame:
        """Get features from the features SQL table.

        This is the preferred method for accessing features for
        offline analysis (assuming data in the local SQL table
        is current).

        Args:
            ticker: Company ticker.
            start: The start date of the observation period. Defaults to the
                first recorded date.
            end: The end date of the observation period. Defaults to the
                last recorded date.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            Fundamental data dataframe with each feature as a
            separate column. Sorted by date.

        Raises:
            `NoResultFound`: If there are no rows for ``ticker`` in the
                refined SQL table.

        Examples:
            >>> finagg.fundam.feat.fundam.normalized.from_refined("AAPL").head(5)  # doctest: +SKIP
                        NORM(PriceBookRatio)  NORM(PriceEarningsRatio)
            date
            2010-01-25             -0.706266                 -0.706279
            2010-01-26             -0.698805                 -0.698935
            2010-01-27             -0.700699                 -0.700799
            2010-01-28             -0.701446                 -0.701534
            2010-01-29             -0.704558                 -0.704598

        """
        start = start or "1776-07-04"
        end = end or utils.today
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.normalized_fundam.name):
            sql.normalized_fundam.create(engine)
        with engine.begin() as conn:
            df = pd.DataFrame(
                conn.execute(
                    sql.normalized_fundam.select().where(
                        sql.normalized_fundam.c.ticker == ticker,
                        sql.normalized_fundam.c.date >= start,
                        sql.normalized_fundam.c.date <= end,
                    )
                )
            )
        if not len(df.index):
            raise NoResultFound(
                f"No industry-normalized fundamental rows found for {ticker}."
            )
        df = df.drop(columns=["ticker"]).set_index("date").sort_index()
        return df

    @classmethod
    def get_candidate_ticker_set(
        cls,
        lb: int = 1,
        *,
        start: None | str = None,
        end: None | str = None,
        engine: None | Engine = None,
    ) -> set[str]:
        """Get all unique tickers in the fundamental SQL table that MAY BE
        ELIGIBLE to be in the feature's SQL table.

        This is just an alias for
        :meth:`finagg.fundam.feat.Fundamental.get_ticker_set`.

        Args:
            lb: Minimum number of rows required to include a ticker in the
                returned set.
            start: The start date of the observation period to include when
                searching for tickers. Defaults to the first recorded date.
            end: The end date of the observation period to include when
                searching for tickers. Defaults to the last recorded date.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            All unique tickers that may be valid for creating
            industry-normalized fundamental features that also have at least
            ``lb`` rows for each tag used for constructing the features.

        Examples:
            >>> "AAPL" in finagg.fundam.feat.fundam.normalized.get_candidate_ticker_set()  # doctest: +SKIP
            True

        """
        return Fundamental.get_ticker_set(lb=lb, start=start, end=end, engine=engine)

    @classmethod
    def get_ticker_set(
        cls,
        lb: int = 1,
        *,
        start: None | str = None,
        end: None | str = None,
        engine: None | Engine = None,
    ) -> set[str]:
        """Get all unique tickers in the feature's SQL table.

        Args:
            lb: Minimum number of rows required to include a ticker in the
                returned set.
            start: The start date of the observation period to include when
                searching for tickers. Defaults to the first recorded date.
            end: The end date of the observation period to include when
                searching for tickers. Defaults to the last recorded date.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            All unique tickers that contain all the columns for creating
            fundamental features that also have at least ``lb`` rows.

        Examples:
            >>> "AAPL" in finagg.fundam.feat.fundam.normalized.get_ticker_set()  # doctest: +SKIP
            True

        """
        start = start or "1776-07-04"
        end = end or utils.today
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.normalized_fundam.name):
            sql.normalized_fundam.create(engine)
        with engine.begin() as conn:
            tickers = (
                conn.execute(
                    sa.select(sql.normalized_fundam.c.ticker)
                    .where(
                        sql.normalized_fundam.c.date >= start,
                        sql.normalized_fundam.c.date <= end,
                    )
                    .group_by(sql.normalized_fundam.c.ticker)
                    .having(sa.func.count(sql.normalized_fundam.c.date) >= lb)
                )
                .scalars()
                .all()
            )
        return set(tickers)

    @classmethod
    def get_tickers_sorted_by(
        cls,
        column: str,
        /,
        *,
        ascending: bool = True,
        date: int | str = 0,
        engine: None | Engine = None,
    ) -> list[str]:
        """Get all tickers in the feature's SQL table sorted by a particular
        column for a date.

        Args:
            column: Feature column to sort by.
            ascending: Whether to return results in ascending order according
                to the values in ``column``.
            date: Date to select from. Defaults to the most recent date with
                available data. If a string, should be in "%Y-%m-%d" format.
                If an integer, should be days before the current date (e.g,
                ``-2`` indicates two days before today).
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            Tickers sorted by a feature column for a particular date.

        Raises:
            `ValueError`: If the integer date is positive.

        Examples:
            >>> ts = finagg.fundam.feat.fundam.normalized.get_tickers_sorted_by(
            ...         "NORM(PriceEarningsRatio)",
            ...         date="2019-01-04"
            ... )
            >>> "AMD" == ts[0]  # doctest: +SKIP
            True

        """
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.normalized_fundam.name):
            sql.normalized_fundam.create(engine)
        with engine.begin() as conn:
            if isinstance(date, int):
                if date > 0:
                    raise ValueError("Date should be non-positive")

                (max_date,) = conn.execute(
                    sa.select(sa.func.max(sql.normalized_fundam.c.date))
                ).one()
                if date == 0:
                    date = str(max_date)
                else:
                    date = (
                        datetime.fromisoformat(str(max_date)) - timedelta(days=date)
                    ).strftime("%Y-%m-%d")

            tickers = (
                conn.execute(
                    sa.select(sql.normalized_fundam.c.ticker)
                    .where(
                        sql.normalized_fundam.c.date == date,
                    )
                    .order_by(sql.normalized_fundam.c[column])
                )
                .scalars()
                .all()
            )
        if not ascending:
            return list(reversed(tickers))
        return list(tickers)

    @classmethod
    def install(
        cls,
        tickers: None | set[str] = None,
        *,
        processes: int = mp.cpu_count() - 1,
        engine: None | Engine = None,
        recreate_tables: bool = False,
    ) -> int:
        """Install data associated with ``tickers`` by pulling data from the
        refined SQL tables, transforming them into normalized features, and
        then writing to the refined normalized fundamental SQL table.

        Tables associated with this method are created if they don't already
        exist.

        Args:
            tickers: Set of tickers to install features for. Defaults to all
                the tickers from :meth:`finagg.fundam.feat.Fundamental.get_ticker_set`.
            processes: Number of background processes to run in parallel
                when processing ``tickers``.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.
            recreate_tables: Whether to drop and recreate tables, wiping all
                previously installed data.

        Returns:
            Number of rows written to the feature's SQL table.

        """
        engine = engine or backend.engine
        tickers = tickers or cls.get_candidate_ticker_set(engine=engine)
        if not tickers:
            logger.info(
                "Skipping finagg.fundam.feat.fundam.normalized installation because no"
                " tickers were provided or no tickers were found with prerequisite data"
                " (i.e., finagg.fundam.feat.fundam data)"
            )
            return 0

        if recreate_tables or not sa.inspect(engine).has_table(
            sql.normalized_fundam.name
        ):
            sql.normalized_fundam.drop(engine, checkfirst=True)
            sql.normalized_fundam.create(engine)

        return utils._install(
            cls.from_other_refined,
            cls.to_refined,
            logger,
            list(tickers),
            engine,
            desc="Installing refined industry-normalized fundamental data",
            processes=processes,
        )

    @classmethod
    def to_refined(
        cls,
        ticker: str,
        df: pd.DataFrame,
        /,
        *,
        engine: None | Engine = None,
    ) -> int:
        """Write the dataframe to the feature store for ``ticker``.

        Args:
            ticker: Company ticker.
            df: Dataframe to store completely as rows in a local SQL
                table.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            Number of rows written to the SQL table.

        Raises:
            `ValueError`: If the given dataframe's columns do not match this
                feature's columns.

        """
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.normalized_fundam.name):
            sql.normalized_fundam.create(engine)
        df = df.reset_index("date")
        df["ticker"] = ticker
        with engine.begin() as conn:
            conn.execute(sql.normalized_fundam.insert(), df.to_dict(orient="records"))  # type: ignore[arg-type]
        return len(df.index)


class Fundamental:
    """Method for gathering fundamental data on a stock using several sources.

    The module variable :data:`finagg.fundam.feat.fundam` is an instance of
    this feature set implementation and is the most popular interface for
    calling feature methods.

    Examples:
        It doesn't matter which data source you use to gather features.
        They all return equivalent dataframes.

        >>> df1 = finagg.fundam.feat.fundam.from_api("AAPL").head(5)
        >>> df2 = finagg.fundam.feat.fundam.from_raw("AAPL").head(5)
        >>> df3 = finagg.fundam.feat.fundam.from_refined("AAPL").head(5)
        >>> pd.testing.assert_frame_equal(df1, df2, rtol=1e-4)
        >>> pd.testing.assert_frame_equal(df1, df3, rtol=1e-4)

    """

    industry = IndustryFundamental()
    """Fundamental features aggregated for an entire industry.
    The most popular way for accessing the :class:`IndustryFundamental`
    feature set.

    :meta hide-value:
    """

    normalized = NormalizedFundamental()
    """A company's fundamental features normalized by its industry.
    The most popular way for accessing the :class:`NormalizedFundamental`
    feature set.

    :meta hide-value:
    """

    @classmethod
    def _normalize(
        cls,
        quarterly: pd.DataFrame,
        prices: pd.DataFrame,
        /,
    ) -> pd.DataFrame:
        """Normalize the feature columns."""
        df = pd.merge(quarterly, prices, how="outer", left_index=True, right_index=True)
        df = df.replace([-np.inf, np.inf], np.nan).ffill()
        df["PriceBookRatio"] = df["close"] / df["BookRatio"]
        df["PriceEarningsRatio"] = df["close"] / df["EarningsPerShareBasic"]
        df.index.names = ["date"]
        df = utils.resolve_col_order(sql.fundam, df)
        return df.dropna()

    @classmethod
    def from_api(
        cls,
        ticker: str,
        /,
        *,
        start: None | str = None,
        end: None | str = None,
    ) -> pd.DataFrame:
        """Get features directly from APIs.

        Args:
            ticker: Company ticker.
            start: The start date of the observation period. Defaults to the
                first recorded date.
            end: The end date of the observation period. Defaults to the
                last recorded date.

        Returns:
            Combined quarterly and daily feature dataframe.
            Sorted by date.

        Examples:
            >>> finagg.fundam.feat.fundam.from_api("AAPL").head(5)  # doctest: +SKIP
                        PriceBookRatio  PriceEarningsRatio
            date
            2010-01-25        0.175061            2.423509
            2010-01-26        0.178035            2.464677
            2010-01-27        0.178813            2.475447
            2010-01-28        0.177153            2.452471
            2010-01-29        0.173825            2.406396

        """
        start = start or "1776-07-04"
        end = end or utils.today
        quarterly = sec.feat.quarterly.from_api(
            ticker,
            start=start,
            end=end,
        ).reset_index(["fy", "fp"], drop=True)
        start = str(quarterly.index[0])
        prices = yfinance.api.get(
            ticker,
            start=start,
            end=end,
        ).set_index("date")
        return cls._normalize(quarterly, prices)

    @classmethod
    def from_raw(
        cls,
        ticker: str,
        /,
        *,
        start: None | str = None,
        end: None | str = None,
        engine: None | Engine = None,
    ) -> pd.DataFrame:
        """Get features directly from other raw SQL tables.

        Args:
            ticker: Company ticker.
            start: The start date of the observation period. Defaults to the
                first recorded date.
            end: The end date of the observation period. Defaults to the
                last recorded date.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            Combined quarterly and daily feature dataframe.
            Sorted by date.

        Examples:
            >>> finagg.fundam.feat.fundam.from_raw("AAPL").head(5)  # doctest: +SKIP
                        PriceBookRatio  PriceEarningsRatio
            date
            2010-01-25        0.175061            2.423509
            2010-01-26        0.178035            2.464677
            2010-01-27        0.178813            2.475447
            2010-01-28        0.177153            2.452471
            2010-01-29        0.173825            2.406396

        """
        start = start or "1776-07-04"
        end = end or utils.today
        engine = engine or backend.engine
        quarterly = sec.feat.quarterly.from_raw(
            ticker,
            start=start,
            end=end,
            engine=engine,
        ).reset_index(["fy", "fp"], drop=True)
        start = str(quarterly.index[0])
        prices = yfinance.feat.prices.from_raw(
            ticker,
            start=start,
            end=end,
            engine=engine,
        )
        return cls._normalize(quarterly, prices)

    @classmethod
    def from_refined(
        cls,
        ticker: str,
        /,
        *,
        start: None | str = None,
        end: None | str = None,
        engine: None | Engine = None,
    ) -> pd.DataFrame:
        """Get features from the feature-dedicated local SQL tables.

        This is the preferred method for accessing features for
        offline analysis (assuming data in the local SQL tables
        is current).

        Args:
            ticker: Company ticker.
            start: The start date of the observation period. Defaults to the
                first recorded date.
            end: The end date of the observation period. Defaults to the
                last recorded date.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            Combined quarterly and daily feature dataframe.
            Sorted by date.

        Raises:
            `NoResultFound`: If there are no rows for ``ticker`` in the
                refined SQL table.

        Examples:
            >>> finagg.fundam.feat.fundam.from_refined("AAPL").head(5)  # doctest: +SKIP
                        PriceBookRatio  PriceEarningsRatio
            date
            2010-01-25        0.175061            2.423509
            2010-01-26        0.178035            2.464677
            2010-01-27        0.178813            2.475447
            2010-01-28        0.177153            2.452471
            2010-01-29        0.173825            2.406396

        """
        start = start or "1776-07-04"
        end = end or utils.today
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.fundam.name):
            sql.fundam.create(engine)
        with engine.begin() as conn:
            df = pd.DataFrame(
                conn.execute(
                    sql.fundam.select().where(
                        sql.fundam.c.ticker == ticker,
                        sql.fundam.c.date >= start,
                        sql.fundam.c.date <= end,
                    )
                )
            )
        if not len(df.index):
            raise NoResultFound(f"No fundamental rows found for {ticker}.")
        df = df.drop(columns=["ticker"]).set_index("date").sort_index()
        return df

    @classmethod
    def get_candidate_ticker_set(
        cls,
        lb: int = 1,
        *,
        start: None | str = None,
        end: None | str = None,
        engine: None | Engine = None,
    ) -> set[str]:
        """Get all unique tickers in the raw SQL table that MAY BE ELIGIBLE
        to be in the feature's SQL table.

        Args:
            lb: Minimum number of rows required to include a ticker in the
                returned set.
            start: The start date of the observation period to include when
                searching for tickers. Defaults to the first recorded date.
            end: The end date of the observation period to include when
                searching for tickers. Defaults to the last recorded date.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            All unique tickers that may be valid for both quarterly and daily
            features that also have at least ``lb`` rows used for constructing
            the features.

        Examples:
            >>> "AAPL" in finagg.fundam.feat.fundam.get_candidate_ticker_set()  # doctest: +SKIP
            True

        """
        return sec.feat.quarterly.get_ticker_set(
            lb=lb, start=start, end=end, engine=engine
        ) & yfinance.feat.daily.get_ticker_set(
            lb=lb, start=start, end=end, engine=engine
        )

    @classmethod
    def get_ticker_set(
        cls,
        lb: int = 1,
        *,
        start: None | str = None,
        end: None | str = None,
        engine: None | Engine = None,
    ) -> set[str]:
        """Get all unique tickers in the feature's SQL table.

        Args:
            lb: Minimum number of rows required to include a ticker in the
                returned set.
            start: The start date of the observation period to include when
                searching for tickers. Defaults to the first recorded date.
            end: The end date of the observation period to include when
                searching for tickers. Defaults to the last recorded date.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            All unique tickers that contain all the columns for creating
            fundamental features that also have at least ``lb`` rows.

        Examples:
            >>> "AAPL" in finagg.fundam.feat.fundam.get_ticker_set()  # doctest: +SKIP
            True

        """
        start = start or "1776-07-04"
        end = end or utils.today
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.fundam.name):
            sql.fundam.create(engine)
        with engine.begin() as conn:
            tickers = (
                conn.execute(
                    sa.select(sql.fundam.c.ticker)
                    .where(
                        sql.fundam.c.date >= start,
                        sql.fundam.c.date <= end,
                    )
                    .group_by(sql.fundam.c.ticker)
                    .having(sa.func.count(sql.fundam.c.date) >= lb)
                )
                .scalars()
                .all()
            )
        return set(tickers)

    @classmethod
    def install(
        cls,
        tickers: None | set[str] = None,
        *,
        processes: int = mp.cpu_count() - 1,
        engine: None | Engine = None,
        recreate_tables: bool = False,
    ) -> int:
        """Install data associated with ``tickers`` by pulling data from other
        refined SQL tables, transforming them into fundamental features, and
        then writing to the refined fundamental features SQL table.

        Tables associated with this method are created if they don't already
        exist.

        Args:
            tickers: Set of tickers to install features for. Defaults to all
                the tickers from :meth:`finagg.sec.feat.Quarterly.get_ticker_set`.
            processes: Number of background processes to run in parallel
                when processing ``tickers``.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.
            recreate_tables: Whether to drop and recreate tables, wiping all
                previously installed data.

        Returns:
            Number of rows written to the feature's SQL table.

        """
        engine = engine or backend.engine
        tickers = tickers or cls.get_candidate_ticker_set(engine=engine)
        if not tickers:
            logger.info(
                "Skipping finagg.fundam.feat.fundam installation because no tickers"
                " were provided or no tickers were found with prerequisite data (i.e.,"
                " finagg.yfinance.feat.daily and finagg.sec.feat.quarterly data)"
            )
            return 0

        if recreate_tables or not sa.inspect(engine).has_table(sql.fundam.name):
            sql.fundam.drop(engine, checkfirst=True)
            sql.fundam.create(engine)

        return utils._install(
            cls.from_raw,
            cls.to_refined,
            logger,
            list(tickers),
            engine,
            desc="Installing refined fundamental data",
            processes=processes,
        )

    @classmethod
    def to_refined(
        cls,
        ticker: str,
        df: pd.DataFrame,
        /,
        *,
        engine: None | Engine = None,
    ) -> int:
        """Write the dataframe to the feature store for ``ticker``.

        Args:
            ticker: Company ticker.
            df: Dataframe to store completely as rows in a local SQL
                table.
            engine: Feature store database engine. Defaults to the engine
                at :data:`finagg.backend.engine`.

        Returns:
            Number of rows written to the SQL table.

        """
        engine = engine or backend.engine
        if not sa.inspect(engine).has_table(sql.fundam.name):
            sql.fundam.create(engine)
        df = df.reset_index("date")
        df["ticker"] = ticker
        with engine.begin() as conn:
            conn.execute(sql.fundam.insert(), df.to_dict(orient="records"))  # type: ignore[arg-type]
        return len(df.index)


fundam = Fundamental()
"""The most popular way for accessing :class:`finagg.fundam.feat.Fundamental`.

:meta hide-value:
"""



import pytest

from finagg.frame import FiscalDelta, FiscalFrame, is_valid_fiscal_seq


@pytest.mark.parametrize(
    "frame,delta,expected",
    [
        (FiscalFrame(2000, 1), FiscalDelta(quarters=1), FiscalFrame(2000, 2)),
        (FiscalFrame(2000, 1), FiscalDelta(quarters=2), FiscalFrame(2000, 3)),
        (FiscalFrame(2000, 1), FiscalDelta(quarters=3), FiscalFrame(2000, 4)),
        (FiscalFrame(2000, 1), FiscalDelta(quarters=4), FiscalFrame(2001, 1)),
        (FiscalFrame(2000, 1), FiscalDelta(quarters=5), FiscalFrame(2001, 2)),
        (FiscalFrame(2000, 4), FiscalDelta(quarters=1), FiscalFrame(2001, 1)),
        (FiscalFrame(2000, 4), FiscalDelta(quarters=2), FiscalFrame(2001, 2)),
        (FiscalFrame(2000, 4), FiscalDelta(quarters=3), FiscalFrame(2001, 3)),
        (FiscalFrame(2000, 4), FiscalDelta(quarters=4), FiscalFrame(2001, 4)),
        (FiscalFrame(2000, 4), FiscalDelta(quarters=5), FiscalFrame(2002, 1)),
    ],
)
def test_fiscal_frame_add_delta(
    frame: FiscalFrame, delta: FiscalDelta, expected: FiscalFrame
) -> None:
    assert (frame + delta) == expected


@pytest.mark.parametrize(
    "frame,delta,expected",
    [
        (FiscalFrame(2000, 1), 1, FiscalFrame(2000, 2)),
        (FiscalFrame(2000, 1), 2, FiscalFrame(2000, 3)),
        (FiscalFrame(2000, 1), 3, FiscalFrame(2000, 4)),
        (FiscalFrame(2000, 1), 4, FiscalFrame(2001, 1)),
        (FiscalFrame(2000, 1), 5, FiscalFrame(2001, 2)),
        (FiscalFrame(2000, 4), 1, FiscalFrame(2001, 1)),
        (FiscalFrame(2000, 4), 2, FiscalFrame(2001, 2)),
        (FiscalFrame(2000, 4), 3, FiscalFrame(2001, 3)),
        (FiscalFrame(2000, 4), 4, FiscalFrame(2001, 4)),
        (FiscalFrame(2000, 4), 5, FiscalFrame(2002, 1)),
    ],
)
def test_fiscal_frame_add_int(
    frame: FiscalFrame, delta: int, expected: FiscalFrame
) -> None:
    assert (frame + delta) == expected


@pytest.mark.parametrize(
    "frame,delta,expected",
    [
        (FiscalFrame(2000, 1), FiscalDelta(quarters=1), FiscalFrame(1999, 4)),
        (FiscalFrame(2000, 1), FiscalDelta(quarters=2), FiscalFrame(1999, 3)),
        (FiscalFrame(2000, 1), FiscalDelta(quarters=3), FiscalFrame(1999, 2)),
        (FiscalFrame(2000, 1), FiscalDelta(quarters=4), FiscalFrame(1999, 1)),
        (FiscalFrame(2000, 1), FiscalDelta(quarters=5), FiscalFrame(1998, 4)),
        (FiscalFrame(2000, 4), FiscalDelta(quarters=1), FiscalFrame(2000, 3)),
        (FiscalFrame(2000, 4), FiscalDelta(quarters=2), FiscalFrame(2000, 2)),
        (FiscalFrame(2000, 4), FiscalDelta(quarters=3), FiscalFrame(2000, 1)),
        (FiscalFrame(2000, 4), FiscalDelta(quarters=4), FiscalFrame(1999, 4)),
        (FiscalFrame(2000, 4), FiscalDelta(quarters=5), FiscalFrame(1999, 3)),
    ],
)
def test_fiscal_frame_sub_delta(
    frame: FiscalFrame, delta: FiscalDelta, expected: FiscalFrame
) -> None:
    assert (frame - delta) == expected


@pytest.mark.parametrize(
    "frame,delta,expected",
    [
        (FiscalFrame(2000, 1), 1, FiscalFrame(1999, 4)),
        (FiscalFrame(2000, 1), 2, FiscalFrame(1999, 3)),
        (FiscalFrame(2000, 1), 3, FiscalFrame(1999, 2)),
        (FiscalFrame(2000, 1), 4, FiscalFrame(1999, 1)),
        (FiscalFrame(2000, 1), 5, FiscalFrame(1998, 4)),
        (FiscalFrame(2000, 4), 1, FiscalFrame(2000, 3)),
        (FiscalFrame(2000, 4), 2, FiscalFrame(2000, 2)),
        (FiscalFrame(2000, 4), 3, FiscalFrame(2000, 1)),
        (FiscalFrame(2000, 4), 4, FiscalFrame(1999, 4)),
        (FiscalFrame(2000, 4), 5, FiscalFrame(1999, 3)),
    ],
)
def test_fiscal_frame_sub_int(
    frame: FiscalFrame, delta: int, expected: FiscalFrame
) -> None:
    assert (frame - delta) == expected


@pytest.mark.parametrize(
    "seq,expected",
    [
        ([1, 1, 2, 1, 1, 2, 1], True),
        ([2, 1, 1, 2, 1, 1], True),
        ([1, 2, 1, 1, 2], True),
        ([2, 2, 1, 1], False),
        ([1, 2, 1, 2, 1, 1], False),
    ],
)
def test_is_valid_fiscal_seq(seq: list[int], expected: bool) -> None:
    assert is_valid_fiscal_seq(seq) == expected



from typing import Callable

import pytest
from requests import HTTPError

import finagg


def skip_if_503(f: Callable) -> None:
    """The BEA API is notorious for being unreliable. We don't really
    support maintaining the BEA API's Python client implementation anymore,
    so we just skip if they error out. We'll keep them just in case we
    feel like supporting it in the future.

    """
    try:
        f()
    except HTTPError as e:
        if e.response.status_code == 503:
            pytest.skip()


def test_fixed_assets_get_parameter_list() -> None:
    skip_if_503(finagg.bea.api.fixed_assets.get_parameter_list)


def test_gdp_by_industry_get_parameter_list() -> None:
    skip_if_503(finagg.bea.api.gdp_by_industry.get_parameter_list)


def test_input_output_get_parameter_list() -> None:
    skip_if_503(finagg.bea.api.input_output.get_parameter_list)


def test_nipa_get_parameter_list() -> None:
    skip_if_503(finagg.bea.api.nipa.get_parameter_list)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import shutil
import sys

# -- Path setup --------------------------------------------------------------

__location__ = os.path.dirname(__file__)

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.join(__location__, "../src"))

# -- Run sphinx-apidoc -------------------------------------------------------
# This hack is necessary since RTD does not issue `sphinx-apidoc` before running
# `sphinx-build -b html . _build/html`. See Issue:
# https://github.com/readthedocs/readthedocs.org/issues/1139
# DON'T FORGET: Check the box "Install your project inside a virtualenv using
# setup.py install" in the RTD Advanced Settings.
# Additionally it helps us to avoid running apidoc manually

try:  # for Sphinx >= 1.7
    from sphinx.ext import apidoc
except ImportError:
    from sphinx import apidoc

output_dir = os.path.join(__location__, "api")
module_dir = os.path.join(__location__, "../src/finagg")
try:
    shutil.rmtree(output_dir)
except FileNotFoundError:
    pass

try:
    import sphinx

    cmd_line = f"sphinx-apidoc --implicit-namespaces -f -o {output_dir} {module_dir}"

    args = cmd_line.split(" ")
    if tuple(sphinx.__version__.split(".")) >= ("1", "7"):
        # This is a rudimentary parse_version to avoid external dependencies
        args = args[1:]

    apidoc.main(args)
except Exception as e:
    print("Running `sphinx-apidoc` failed!\n{}".format(e))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "finagg"
copyright = "Andrew B"
author = "Andrew B"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.autosummary",
    "sphinx.ext.viewcode",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.ifconfig",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "sphinx.ext.extlinks",
    "sphinx_copybutton",
    "sphinx_rtd_theme",
    "sphinxcontrib.programoutput",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "build", "Thumbs.db", ".DS_Store"]
nitpicky = True
nitpick_ignore = [
    ("py:class", "requests.models.Response"),
    ("py:class", "requests.exceptions.RequestException"),
    ("py:class", "sqlalchemy.sql.schema.MetaData"),
    ("py:class", "sqlalchemy.sql.schema.Table"),
    ("py:class", "sqlalchemy.engine.base.Engine"),
    ("py:class", "_P"),
    ("py:exc", "NoResultFound"),
    ("py:obj", "finagg.ratelimit._P"),
]

# -- Autodoc options ---------------------------------------------------------

autodoc_member_order = "bysource"

# -- doctest options ---------------------------------------------------------

doctest_global_setup = """
import finagg
import pandas as pd
pd.set_option("display.precision", 4)
"""

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]
html_css_files = [
    "custom.css",  # Avoid name clashes with the theme
]

# -- External mapping --------------------------------------------------------

python_version = ".".join(map(str, sys.version_info[0:2]))
intersphinx_mapping = {
    "python": ("https://docs.python.org/" + python_version, None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable", None),
    "sqlalchemy": ("https://docs.sqlalchemy.org/en/20/", None),
}



"""Customizable rate-limiting for requests-style getters.

The definitions within this submodule are used throughout :mod:`finagg` for
respecting 3rd party API rate limits to avoid server-side throttling.

"""

import time
from abc import ABC, abstractmethod
from collections import deque
from dataclasses import dataclass
from datetime import timedelta
from functools import update_wrapper
from typing import Callable, Generic, ParamSpec, Sequence

import requests

_P = ParamSpec("_P")


@dataclass
class _RateLimitData:
    """Convenience container for rate limit data."""

    #: Timestamp associated with the data.
    ts: float

    #: Response's contribution to a rate limit threshold.
    quantity: float


@dataclass
class _RunningTotals:
    """Track running totals associated with exceeding rate limits."""

    #: Total limit contribution of all responses within a window towards
    #: a rate limit threshold.
    quantity: float = 0.0

    #: Max wait time needed until another request can be made without
    #: exceeding a rate limit.
    wait: float = 0.0


class RateLimit(ABC):
    """Interface for defining a rate limit for an external API getter.

    You can create a custom rate-limiter by inheriting from this class
    and implementing a custom :meth:`eval` method.

    Args:
        limit: Max limit within ``period`` (e.g., max number of
            requests, errors, size in memory, etc.).
        period: Time interval for evaluating ``limit``.
        buffer: Reduce ``limit`` by this fraction. Adds a bit of
            leeway to ensure ``limit`` is not reached. Useful for
            enforcing response size limits.

    .. seealso::
        :meth:`guard`: For the intended usage of getting a
            :class:`RateLimitGuard` instance.
        :class:`RequestLimit`: For an example of a request
            rate limiter.

    """

    #: Deque of responses containing limits from ``eval`` and
    #: the timesteps they were observed.
    _rate_limit_datas: deque[_RateLimitData]

    #: Running total contributions to ``limit`` within ``period``
    #: and the total wait time.
    _running_totals: _RunningTotals

    #: Max quantity allowed within ``period``. The quantity type being limited
    #: is dependent on what's returned by :meth:`eval`.
    limit: float

    #: Time interval for evaluating ``limit`` (in seconds).
    period: float

    def __init__(
        self, limit: float, period: float | timedelta, /, *, buffer: float = 0.0
    ) -> None:
        self.limit = limit * (1 - buffer)
        self.period = (
            period.total_seconds() if isinstance(period, timedelta) else period
        )
        self._rate_limit_datas = deque()
        self._running_totals = _RunningTotals()

    @property
    def _ts(self) -> float:
        """Get the most recent response timestamp as a result of calling
        the underlying getter.

        """
        return (
            self._rate_limit_datas[-1].ts
            if self._rate_limit_datas
            else time.perf_counter()
        )

    def _update(self, response: requests.Response, /) -> float:
        """Update the rate limit's running ``total`` and ``responses``
        collection.

        This method calls :meth:`eval` and uses its return value to
        update the client-side throttling wait time (if there is any) to
        avoid exceeding server-side API limits.

        Args:
            response: Request response (possibly cached).

        Returns:
            Estimated time to wait to avoid being throttled.

        """
        # Get the timestamp and timedelta associated with this response.
        ts = time.perf_counter()
        dt = max(ts - self._ts, 0.0)

        # Get the limit contribution value and wait time from the response.
        v = self.eval(response)
        if not isinstance(v, dict):
            v = {"limit": v}
        if "wait" not in v:
            v["wait"] = 0.0
        quantity = v["limit"]
        wait = v["wait"]

        # Create the new rate limit data.
        new = _RateLimitData(ts, quantity)
        self._rate_limit_datas.append(new)

        # Update running totals for estimating wait times.
        self._running_totals.quantity += quantity
        self._running_totals.wait = max(self._running_totals.wait - dt, 0.0)
        self._running_totals.wait = max(self._running_totals.wait, wait)

        # Remove timed-out responses and remove their contributions
        # to the total limit.
        while self._rate_limit_datas and (
            (self._rate_limit_datas[-1].ts - self._rate_limit_datas[0].ts)
            >= self.period
        ):
            old = self._rate_limit_datas.popleft()
            self._running_totals.quantity -= old.quantity

        # Update total wait time according to the max limit.
        tmp_limit = self._running_totals.quantity
        tmp_wait = 0.0
        if self._running_totals.quantity >= self.limit:
            for r in self._rate_limit_datas:
                tmp_wait = self.period - (self._ts - r.ts)
                tmp_limit -= r.quantity
                if tmp_limit < self.limit:
                    break

        self._running_totals.wait = max(self._running_totals.wait, tmp_wait)
        return self._running_totals.wait

    @abstractmethod
    def eval(self, response: requests.Response, /) -> float | dict[str, float]:
        """Evaluate a response and determine how much it contributes
        to the max limit imposed by this instance.

        This is the main method that should be overwritten by subclasses
        to create custom rate-limiters. This method is called with each
        requests's response to determine how much that request/response
        contributes to the rate-limiting.

        Args:
            response: Request response (possibly cached).

        Returns:
            A number indicating the request/response's contribution
            to the rate limit OR a dictionary containing:

                - "limit": a number indicating the request/response's
                  contribution to the rate limit
                - "wait": time to wait before a new request can be made

        """


class RequestLimit(RateLimit):
    """Limit the number of requests made by the underlying getter."""

    def eval(self, response: requests.Response, /) -> float | dict[str, float]:
        if hasattr(response, "from_cache") and response.from_cache:
            return 0.0
        return float(1)


class ErrorLimit(RateLimit):
    """Limit the number of errors occurred when using the underlying getter."""

    def eval(self, response: requests.Response, /) -> float | dict[str, float]:
        if hasattr(response, "from_cache") and response.from_cache:
            return 0.0
        return float(response.status_code != 200)


class SizeLimit(RateLimit):
    """Limit the size of responses when using the underlying getter."""

    def eval(self, response: requests.Response, /) -> float | dict[str, float]:
        if hasattr(response, "from_cache") and response.from_cache:
            return 0.0
        return float(len(response.content))


class RateLimitGuard(Generic[_P]):
    """Wraps requests-like getters to introduce blocking functionality
    when requests are getting close to violating call limits.

    Args:
        f: Requests-style getter that's wrapped and rate-limited.
        limits: Limits to apply to the requests-style getter.
        warn: Whether to print a message to stdout whenever client-side
            throttling is occurring to respect ``limits``.

    .. seealso::
        :meth:`guard`: For the intended usage of getting a
            :class:`RateLimitGuard` instance.
        :class:`RequestLimit`: For an example of a request
            rate limiter.

    """

    #: ``requests``-like getter that returns a response.
    f: Callable[_P, requests.Response]

    #: Limits to apply to requests/responses.
    limits: tuple[RateLimit, ...]

    #: Whether to print a warning when requests are being temporarily blocked
    #: to respect imposed rate limits.
    warn: bool

    def __init__(
        self,
        f: Callable[_P, requests.Response],
        limits: tuple[RateLimit, ...],
        /,
        *,
        warn: bool = False,
    ) -> None:
        self.f = f
        self.limits = limits
        self.warn = warn
        update_wrapper(self, f)

    def __call__(self, *args: _P.args, **kwargs: _P.kwargs) -> requests.Response:
        """Call the underlying getter and sleep the wait required to
        satisfy the guard's limits.

        Args:
            *args: Args passed to the underlying getter.
            **kwargs: Kwargs passed to the underlying getter.

        Returns:
            The received response.

        """
        r = self.f(*args, **kwargs)
        wait = 0.0
        for limit in self.limits:
            tmp_wait = limit._update(r)
            wait = max(wait, tmp_wait)
        if wait > 0:
            if self.warn:
                print(f"Throttling requests to {r.url} for {wait:.2f} (s)", flush=True)
            time.sleep(wait)
        return r


def guard(
    limits: Sequence[RateLimit], /, *, warn: bool = False
) -> Callable[[Callable[_P, requests.Response],], RateLimitGuard[_P]]:
    """Apply ``limits`` to a requests-style getter.

    Args:
        limits: Rate limits to apply to the requests-style getter.
        warn: Whether to print a message when client-side throttling is
            occurring.

    Returns:
        A decorator that wraps the original requests-style getter in a
        :class:`RateLimitGuard` to avoid exceeding ``limits``.

    Examples:
        Limit 5 requests to Google per second.

        >>> import requests
        >>> from datetime import timedelta
        >>> from finagg.ratelimit import RequestLimit, guard
        >>> @guard([RequestLimit(5, timedelta(seconds=1))])
        ... def get() -> requests.Response:
        ...     return requests.get("https://google.com")

    """

    def decorator(f: Callable[_P, requests.Response], /) -> RateLimitGuard[_P]:
        return RateLimitGuard(f, tuple(limits), warn=warn)

    return decorator



"""The "fred/release" and "fred/releases" API.

See the official FRED API docs for more info:

    https://fred.stlouisfed.org/docs/api/fred/

"""

import pandas as pd

from . import _api


class ReleasesDates(_api.API):
    """Get all release dates of FRED economic data.

    The class variable :attr:`finagg.fred.api.Releases.dates` is an instance
    of this API implementation and is the most popular interface for calling
    this API.

    """

    url = "https://api.stlouisfed.org/fred/releases/dates"

    @classmethod
    def get(
        cls,
        *,
        realtime_start: None | int | str = None,
        realtime_end: None | int | str = None,
        limit: None | int = 1000,
        offset: None | int = 0,
        order_by: None | str = "release_date",
        sort_order: None | str = "desc",
        include_release_dates_with_no_data: None | bool = False,
        paginate: bool = False,
        cache: bool = True,
        api_key: None | str = None,
    ) -> pd.DataFrame:
        """Get all release dates of economic data.

        See the related FRED API documentation at:

            https://fred.stlouisfed.org/docs/api/fred/releases_dates.html

        Args:
            realtime_start: Start date for fetching results
                according to their publication date.
            realtime_end: End date for fetching results according
                to their publication date.
            limit: Maximum number of results to return.
            offset: Result start offset.
            order_by: Variable to order results by. Options include:

                - "release_date"
                - "release_id"
                - "release_name"

            sort_order: Sort results in ascending ("asc") or
                descending ("desc") order.
            include_release_dates_with_no_data: Whether to return release
                dates that don't contain any data.
            paginate: Whether to manage `offset` automatically, making multiple
                API calls until all results are returned.
            cache: Whether to cache the response from the API.
            api_key: Your FRED API key. Defaults to the ``FRED_API_KEY``
                environment variable.

        Returns:
            A dataframe containing data on release dates for all
            releases of economic data.

        """
        return _api.maybe_paginate(
            "release_dates",
            cls.url,
            realtime_start=realtime_start,
            realtime_end=realtime_end,
            limit=limit,
            offset=offset,
            order_by=order_by,
            sort_order=sort_order,
            include_release_dates_with_no_data=include_release_dates_with_no_data,
            paginate=paginate,
            cache=cache,
            api_key=api_key,
        )


class Releases(_api.API):
    """Get all releases of economic data."""

    dates = ReleasesDates()
    """"releases/dates" FRED API. Get dates for releases of economic data.
    The most popular way for accessing the
    :class:`finagg.fred.api.ReleasesDates` API.

    :meta hide-value:
    """

    url = "https://api.stlouisfed.org/fred/releases"

    @classmethod
    def get(
        cls,
        *,
        realtime_start: None | int | str = None,
        realtime_end: None | int | str = None,
        limit: None | int = 1000,
        offset: None | int = 0,
        order_by: None | str = None,
        sort_order: None | str = None,
        paginate: bool = False,
        cache: bool = True,
        api_key: None | str = None,
    ) -> pd.DataFrame:
        """Get all releases of economic data.

        See the related FRED API documentation at:

            https://fred.stlouisfed.org/docs/api/fred/releases.html

        Args:
            realtime_start: Start date for fetching results
                according to their publication date.
            realtime_end: End date for fetching results according
                to their publication date.
            limit: Maximum number of results to return.
            offset: Result start offset.
            order_by: Variable to order results by. Options include:

                - "release_id"
                - "name"
                - "press_release"
                - "realtime_start"
                - "realtime_end"

            sort_order: Sort results in ascending ("asc") or
                descending ("desc") order.
            paginate: Whether to manage `offset` automatically, making multiple
                API calls until all results are returned.
            cache: Whether to cache the response from the API.
            api_key: Your FRED API key. Defaults to the ``FRED_API_KEY``
                environment variable.

        Returns:
            A dataframe containing data on all releases of economic
            data.

        """
        return _api.maybe_paginate(
            "releases",
            cls.url,
            realtime_start=realtime_start,
            realtime_end=realtime_end,
            limit=limit,
            offset=offset,
            order_by=order_by,
            sort_order=sort_order,
            paginate=paginate,
            cache=cache,
            api_key=api_key,
        )


class ReleaseDates(_api.API):
    """Get dates associated with an economic release.

    The class variable :attr:`finagg.fred.api.Release.dates` is an
    instance of this API implementation and is the most popular interface for
    calling this API.

    """

    url = "https://api.stlouisfed.org/fred/release/dates"

    @classmethod
    def get(
        cls,
        release_id: int,
        /,
        *,
        realtime_start: None | int | str = None,
        realtime_end: None | int | str = None,
        limit: None | int = 10000,
        offset: None | int = 0,
        sort_order: None | str = None,
        include_release_dates_with_no_data: None | bool = False,
        paginate: bool = False,
        cache: bool = True,
        api_key: None | str = None,
    ) -> pd.DataFrame:
        """Get data on release dates for a particular release of economic data.

        See the related FRED API documentation at:

            https://fred.stlouisfed.org/docs/api/fred/release_dates.html

        Args:
            release_id: The ID for a release.
            realtime_start: Start date for fetching results
                according to their publication date.
            realtime_end: End date for fetching results according
                to their publication date.
            limit: Maximum number of results to return.
            offset: Result start offset.
            sort_order: Sort results in ascending ("asc") or
                descending ("desc") order.
            include_release_dates_with_no_data: Whether to return release
                dates that don't contain any data.
            paginate: Whether to manage `offset` automatically, making multiple
                API calls until all results are returned.
            cache: Whether to cache the response from the API.
            api_key: Your FRED API key. Defaults to the ``FRED_API_KEY``
                environment variable.

        Returns:
            A dataframe containing data for an economic data release's release dates.

        """
        return _api.maybe_paginate(
            "release_dates",
            cls.url,
            release_id=release_id,
            realtime_start=realtime_start,
            realtime_end=realtime_end,
            limit=limit,
            offset=offset,
            sort_order=sort_order,
            include_release_dates_with_no_data=include_release_dates_with_no_data,
            paginate=paginate,
            cache=cache,
            api_key=api_key,
        )


class ReleaseSeries(_api.API):
    """Get series associated with an economic release.

    The class variable :attr:`finagg.fred.api.Release.series` is an
    instance of this API implementation and is the most popular interface for
    calling this API.

    """

    url = "https://api.stlouisfed.org/fred/release/series"

    @classmethod
    def get(
        cls,
        release_id: int,
        /,
        *,
        realtime_start: None | int | str = None,
        realtime_end: None | int | str = None,
        limit: None | int = 1000,
        offset: None | int = 0,
        order_by: None | str = None,
        sort_order: None | str = None,
        filter_variable: None | str = None,
        filter_value: None | str = None,
        tag_names: None | str | list[str] = None,
        exclude_tag_names: None | str | list[str] = None,
        paginate: bool = False,
        cache: bool = True,
        api_key: None | str = None,
    ) -> pd.DataFrame:
        """Get data on the series related to a release of economic data.

        See the related FRED API documentation at:

            https://fred.stlouisfed.org/docs/api/fred/release_series.html

        Args:
            release_id: The ID for a release.
            realtime_start: Start date for fetching results
                according to their publication date.
            realtime_end: End date for fetching results according
                to their publication date.
            limit: Maximum number of results to return.
            offset: Result start offset.
            order_by: Variable to order results by. Options include:

                - "series_id"
                - "title"
                - "units"
                - "frequency"
                - "seasonal_adjustment"
                - "realtime_start"
                - "realtime_end"
                - "last_updated"
                - "observation_start"
                - "observation_end"
                - "popularity"
                - "group_popularity"

            sort_order: Sort results in ascending ("asc") or
                descending ("desc") order.
            filter_variable: The attribute (or column) to filter results by.
                Options include:

                - "frequency"
                - "units"
                - "seasonal_adjustment"

            filter_value: The value of `filter_variable` to filter results
                by.
            tag_names: Find tags related to these tags.
            exclude_tag_names: Exclude tags related to these tags.
            paginate: Whether to manage `offset` automatically, making multiple
                API calls until all results are returned.
            cache: Whether to cache the response from the API.
            api_key: Your FRED API key. Defaults to the ``FRED_API_KEY``
                environment variable.

        Returns:
            A dataframe containing series data for a release.

        """
        return _api.maybe_paginate(
            "seriess",
            cls.url,
            release_id=release_id,
            realtime_start=realtime_start,
            realtime_end=realtime_end,
            limit=limit,
            offset=offset,
            order_by=order_by,
            sort_order=sort_order,
            filter_variable=filter_variable,
            filter_value=filter_value,
            tag_names=tag_names,
            exclude_tag_names=exclude_tag_names,
            paginate=paginate,
            cache=cache,
            api_key=api_key,
        )


class ReleaseSources(_api.API):
    """Get sources associated with an economic release.

    The class variable :attr:`finagg.fred.api.Release.sources` is an
    instance of this API implementation and is the most popular interface for
    calling this API.

    """

    url = "https://api.stlouisfed.org/fred/release/sources"

    @classmethod
    def get(
        cls,
        release_id: int,
        /,
        *,
        realtime_start: None | int | str = None,
        realtime_end: None | int | str = None,
        cache: bool = True,
        api_key: None | str = None,
    ) -> pd.DataFrame:
        """Get sources related to an economic release.

        See the related FRED API documentation at:

            https://fred.stlouisfed.org/docs/api/fred/release_sources.html

        Args:
            release_id: The ID for a release.
            realtime_start: Start date for fetching results
                according to their publication date.
            realtime_end: End date for fetching results according
                to their publication date.
            cache: Whether to cache the response from the API.
            api_key: Your FRED API key. Defaults to the ``FRED_API_KEY``
                environment variable.

        Returns:
            A dataframe containing sources related to an economic release.

        """
        data = _api.get(
            cls.url,
            release_id=release_id,
            realtime_start=realtime_start,
            realtime_end=realtime_end,
            cache=cache,
            api_key=api_key,
        ).json()
        data = data["sources"]
        return pd.DataFrame(data)


class ReleaseTags(_api.API):
    """Get tags for an economic release.

    The class variable :attr:`finagg.fred.api.Release.tags` is an
    instance of this API implementation and is the most popular interface for
    calling this API.

    """

    url = "https://api.stlouisfed.org/fred/release/tags"

    @classmethod
    def get(
        cls,
        release_id: int,
        /,
        *,
        realtime_start: None | int | str = None,
        realtime_end: None | int | str = None,
        tag_names: None | str | list[str] = None,
        tag_group_id: None | str = None,
        search_text: None | str | list[str] = None,
        limit: None | int = 1000,
        offset: None | int = 0,
        order_by: None | str = None,
        sort_order: None | str = None,
        paginate: bool = False,
        cache: bool = True,
        api_key: None | str = None,
    ) -> pd.DataFrame:
        """Get tags for an economic release.

        See the related FRED API documentation at:

            https://fred.stlouisfed.org/docs/api/fred/release_tags.html

        Args:
            release_id: The ID for a release.
            realtime_start: Start date for fetching results
                according to their publication date.
            realtime_end: End date for fetching results according
                to their publication date.
            tag_names: Filtering of tag names to include in the results.
            tag_group_id: A tag group ID to filter tags by. Options include:

                - "freq" = frequency
                - "gen" = general or concept
                - "geo" = geography
                - "geot" = geography type
                - "rls" = release
                - "seas" = seasonal adjustment
                - "src" = source

            search_text: The words to find matching tags with.
            limit: Maximum number of results to return.
            offset: Result start offset.
            order_by: Variable to order results by. Options include:

                - "series_count"
                - "popularity"
                - "created"
                - "name"
                - "group_id"

            sort_order: Sort results in ascending ("asc") or
                descending ("desc") order.
            paginate: Whether to manage `offset` automatically, making multiple
                API calls until all results are returned.
            cache: Whether to cache the response from the API.
            api_key: Your FRED API key. Defaults to the ``FRED_API_KEY``
                environment variable.

        Returns:
            A dataframe containing data for an economic release's tags
            according to the given parameters.

        """
        return _api.maybe_paginate(
            "tags",
            cls.url,
            release_id=release_id,
            realtime_start=realtime_start,
            realtime_end=realtime_end,
            tag_names=tag_names,
            tag_group_id=tag_group_id,
            search_text=search_text,
            limit=limit,
            offset=offset,
            order_by=order_by,
            sort_order=sort_order,
            paginate=paginate,
            cache=cache,
            api_key=api_key,
        )


class ReleaseRelatedTags(_api.API):
    """Get tags related to an economic release.

    The class variable :attr:`finagg.fred.api.Release.related_tags` is an
    instance of this API implementation and is the most popular interface for
    calling this API.

    """

    url = "https://api.stlouisfed.org/fred/release/related_tags"

    @classmethod
    def get(
        cls,
        release_id: int,
        /,
        *,
        realtime_start: None | int | str = None,
        realtime_end: None | int | str = None,
        tag_names: None | str | list[str] = None,
        exclude_tag_names: None | str | list[str] = None,
        tag_group_id: None | str = None,
        search_text: None | str | list[str] = None,
        limit: None | int = 1000,
        offset: None | int = 0,
        order_by: None | str = None,
        sort_order: None | str = None,
        paginate: bool = False,
        cache: bool = True,
        api_key: None | str = None,
    ) -> pd.DataFrame:
        """Get data for tags related to an economic release.

        See the related FRED API documentation at:

            https://fred.stlouisfed.org/docs/api/fred/release_related_tags.html

        Args:
            release_id: The ID for a release.
            realtime_start: Start date for fetching results
                according to their publication date.
            realtime_end: End date for fetching results according
                to their publication date.
            tag_names: Find tags related to these tags.
            exclude_tag_names: Exclude tags related to these tags.
            tag_group_id: A tag group ID to filter tags by. Options include:

                - "freq" = frequency
                - "gen" = general or concept
                - "geo" = geography
                - "geot" = geography type
                - "rls" = release
                - "seas" = seasonal adjustment
                - "src" = source

            search_text: The words to find matching tags with.
            limit: Maximum number of results to return.
            offset: Result start offset.
            order_by: Variable to order results by. Options include:

                - "series_count"
                - "popularity"
                - "created"
                - "name"
                - "group_id"

            sort_order: Sort results in ascending ("asc") or
                descending ("desc") order.
            paginate: Whether to manage `offset` automatically, making multiple
                API calls until all results are returned.
            cache: Whether to cache the response from the API.
            api_key: Your FRED API key. Defaults to the ``FRED_API_KEY``
                environment variable.

        Returns:
            A dataframe containing data for tags related to an economic
            release according to the given parameters.

        """
        return _api.maybe_paginate(
            "tags",
            cls.url,
            release_id=release_id,
            realtime_start=realtime_start,
            realtime_end=realtime_end,
            tag_names=tag_names,
            exclude_tag_names=exclude_tag_names,
            tag_group_id=tag_group_id,
            search_text=search_text,
            limit=limit,
            offset=offset,
            order_by=order_by,
            sort_order=sort_order,
            paginate=paginate,
            cache=cache,
            api_key=api_key,
        )


class ReleaseTables(_api.API):
    """Get tables associated with an economic release.

    The class variable :attr:`finagg.fred.api.Release.tables` is an instance of
    this API implementation and is the most popular interface for calling this
    API.

    """

    url = "https://api.stlouisfed.org/fred/release/tables"

    @classmethod
    def get(
        cls,
        release_id: int,
        /,
        *,
        element_id: None | int = 0,
        include_observation_values: None | bool = False,
        observation_date: None | str = None,
        cache: bool = True,
        api_key: None | str = None,
    ) -> pd.DataFrame:
        """Get release tables for a given economic release.

        See the related FRED API documentation at:

            https://fred.stlouisfed.org/docs/api/fred/release_tables.html

        Args:
            release_id: The ID for a release.
            element_id: The release table element you'd like to retrieve.
            include_observation_values: A flag indicating whether observations
                need to be returned.
            observation_date: The observation date to be included with the
                returned release table.
            cache: Whether to cache the response from the API.
            api_key: Your FRED API key. Defaults to the ``FRED_API_KEY``
                environment variable.

        Returns:
            A dataframe of release tables for a given economic release.

        """
        data = _api.get(
            cls.url,
            release_id=release_id,
            element_id=element_id,
            include_observation_values=include_observation_values,
            observation_date=observation_date,
            cache=cache,
            api_key=api_key,
        ).json()
        data = data["tables"]
        return pd.DataFrame(data)


class Release(_api.API):
    """Get data on an economic release.

    The class variable :data:`finagg.fred.api.release` is an instance of this
    API implementation and is the most popular interface for calling this API.

    """

    dates = ReleaseDates()
    """"release/dates" FRED API. Get economic release dates.
    The most popular way for accessing the
    :class:`finagg.fred.api.ReleaseDates` API.

    :meta hide-value:
    """

    related_tags = ReleaseRelatedTags()
    """"release/related_tags" FRED API. Get tags related to an economic release.
    The most popular way for accessing the
    :class:`finagg.fred.api.ReleaseRelatedTags` API.

    :meta hide-value:
    """

    series = ReleaseSeries()
    """"release/series" FRED API. Get the series of an economic release.
    The most popular way for accessing the
    :class:`finagg.fred.api.ReleaseSeries` API.

    :meta hide-value:
    """

    sources = ReleaseSources()
    """"release/sources" FRED API. Get the sources for an economic release.
    The most popular way for accessing the
    :class:`finagg.fred.api.ReleaseSources` API.

    :meta hide-value:
    """

    tables = ReleaseTables()
    """"release/tables" FRED API. Get the tables of an economic release.
    The most popular way for accessing the
    :class:`finagg.fred.api.ReleaseTables` API.

    :meta hide-value:
    """

    tags = ReleaseTags()
    """"release/tags" FRED API. Get tags of an economic release.
    The most popular way for accessing the
    :class:`finagg.fred.api.ReleaseTags` API.

    :meta hide-value:
    """

    url = "https://api.stlouisfed.org/fred/release"

    @classmethod
    def get(
        cls,
        release_id: int,
        /,
        *,
        realtime_start: None | int | str = None,
        realtime_end: None | int | str = None,
        cache: bool = True,
        api_key: None | str = None,
    ) -> pd.DataFrame:
        """Get overview data of an economic release.

        See the related FRED API documentation at:

            https://fred.stlouisfed.org/docs/api/fred/release.html

        Args:
            release_id: The ID for a release.
            realtime_start: Start date for fetching results
                according to their publication date.
            realtime_end: End date for fetching results according
                to their publication date.
            cache: Whether to cache the response from the API.
            api_key: Your FRED API key. Defaults to the ``FRED_API_KEY``
                environment variable.

        Returns:
            A dataframe containing high-level info on an economic release.

        """
        data = _api.get(
            cls.url,
            release_id=release_id,
            realtime_start=realtime_start,
            realtime_end=realtime_end,
            cache=cache,
            api_key=api_key,
        ).json()
        data = data["releases"]
        return pd.DataFrame(data)
